# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
import pytest
import math
import pickle

# Import from 'ezdxf.math._vector' to test Python implementation
from ezdxf.math._vector import Vec2, Vec3
from ezdxf.acc import USE_C_EXT

all_vec_classes = [Vec2, Vec3]
vec2_only = [Vec2]

if USE_C_EXT:
    from ezdxf.acc.vector import Vec2 as CVec2

    all_vec_classes.append(CVec2)
    vec2_only.append(CVec2)


# Vec2 is a sub set of Vec3, Vec3 can do everything Vec2 can do, but not every
# operation has the same result for 2D and 3D.
@pytest.fixture(params=all_vec_classes)
def vcls(request):
    return request.param


@pytest.fixture(params=vec2_only)
def vec2(request):
    return request.param


def test_init_tuple(vcls):
    v = vcls((2, 3))
    assert v.x == 2
    assert v.y == 3


def test_empty_init(vcls):
    v = vcls()
    assert v.x == 0.
    assert v.y == 0.


def test_init_vec2(vcls):
    v = Vec2(vcls(2, 3))
    assert v.x == 2
    assert v.y == 3


def test_compatible_to_vector():
    v = Vec3(Vec2(1, 2))
    assert v == (1, 2, 0)

    v = Vec2(Vec3(1, 2, 3))
    assert v.x == 1
    assert v.y == 2


def test_vec3(vec2):
    v = vec2(1, 2)
    assert len(v) == 2
    v3 = v.vec3
    assert len(v3) == 3
    assert v3 == (1, 2, 0)


def test_round(vec2):
    v = vec2(1.123, 2.123)
    v2 = v.round(1)
    assert v2 == (1.1, 2.1)


def test_from_angle(vcls):
    angle = math.radians(50)
    length = 3.0
    assert vcls.from_angle(angle, length) == vcls(
        (math.cos(angle) * length, math.sin(angle) * length)
    )


def test_vec2_as_tuple(vec2):
    v = vec2(1, 2)
    assert v[0] == 1
    assert v[1] == 2

    with pytest.raises(IndexError):
        _ = v[2]
    # negative indices not supported
    with pytest.raises(IndexError):
        _ = v[-1]


def test_iter(vcls):
    assert sum(vcls(1, 2)) == 3


def test_deep_copy():
    import copy

    v = Vec2(1, 2)
    l1 = [v, v, v]
    l2 = copy.copy(l1)
    assert l2[0] is l2[1]
    assert l2[1] is l2[2]
    assert l2[0] is v

    # Vec3, CVec2 and CVec3 are immutable and do not create copies of itself!
    l3 = copy.deepcopy(l1)
    assert l3[0] is l3[1]
    assert l3[1] is l3[2]
    assert l3[0] is not v


def test_get_angle(vcls):
    v = vcls(3, 3)
    assert math.isclose(v.angle_deg, 45)
    assert math.isclose(v.angle, math.radians(45))


def test_compare_vectors(vcls):
    v1 = vcls(1, 2)
    assert v1 == v1

    v2 = vcls(2, 3)
    assert v2 > v1
    assert v1 < v2


def test_is_close(vcls):
    v1 = vcls(421846.9857097387, -36908.41493252139)
    v2 = vcls(421846.9857097387, -36908.41493252141)
    assert v1.isclose(v2) is True


def test_is_null(vcls):
    v = vcls(0, 0)
    assert v.is_null is True

    v1 = vcls(23.56678, 56678.56778) * (1.0 / 14.5667)
    v2 = vcls(23.56678, 56678.56778) / 14.5667
    assert (v2 - v1).is_null


def test_is_not_null_default_abs_tol(vcls):
    assert vcls(1e-11, 0).is_null is False


def test_is_null_default_abs_tol(vcls):
    assert vcls(1e-12, 0).is_null is True


def test_bool(vcls):
    v = vcls((0, 0))
    assert bool(v) is False

    v1 = vcls(23.56678, 56678.56778) * (1.0 / 14.5667)
    v2 = vcls(23.56678, 56678.56778) / 14.5667
    result = v2 - v1
    assert bool(result) is False
    # current rel_tol=1e-9
    assert not vcls(1e-8, 0).is_null


def test_magnitude(vcls):
    v = vcls(3, 4)
    assert math.isclose(abs(v), 5)
    assert math.isclose(v.magnitude, 5)


def test_normalize(vcls):
    v = vcls(2, 0)
    assert v.normalize() == (1, 0)


def test_normalize_to_length(vcls):
    v = vcls(2, 0)
    assert v.normalize(4) == (4, 0)


def test_orthogonal_ccw(vcls):
    v = vcls(3, 4)
    assert v.orthogonal() == (-4, 3)


def test_orthogonal_cw(vcls):
    v = vcls(3, 4)
    assert v.orthogonal(False) == (4, -3)


def test_negative(vcls):
    v = vcls(2, 3)
    assert -v == (-2, -3)


def test_add_vector(vcls):
    assert vcls(2, 3) + vcls(7, 7) == (9, 10)


def test_add_vec3(vec2):
    assert vec2(2, 3) + Vec3(7, 7) == (9, 10)


def test_iadd_vector(vec2):
    v = Vec2(2, 3)
    v += Vec2(7, 7)
    assert v == (9, 10)


def test_inplace_operations_do_not_mutate_vec2_inplace():
    v = Vec2(2, 3)
    v_check = v
    v += Vec2(7, 7)
    assert v_check is not v, "__iadd__ should not operate inplace"

    v_check = v
    v -= Vec2(7, 7)
    assert v_check is not v, "__isub__ should not operate inplace"

    v_check = v
    v *= 1
    assert v_check is not v, "__imul__ should not operate inplace"

    v_check = v
    v /= 1
    assert v_check is not v, "__itruediv__ should not operate inplace"


def test_add_scalar_type_erorr(vcls):
    with pytest.raises(TypeError):
        vcls(1, 1) + 1


def test_iadd_scalar_type_error(vcls):
    v = vcls(2, 3)
    with pytest.raises(TypeError):
        v += 1


def test_radd_scalar_type_error(vcls):
    with pytest.raises(TypeError):
        1 + vcls(1, 1)


def test_radd_tuple_type_error(vec2):
    with pytest.raises(TypeError):
        (1, 1) + vec2(1, 1)


def test_sub_vector(vcls):
    assert vcls(2, 3) - vcls(7, 7) == (-5, -4)


def test_isub_vector(vec2):
    v = Vec2(2, 3)
    v -= Vec2(7, 7)
    assert v == (-5, -4)


def test_sub_vec3(vec2):
    assert vec2(2, 3) - Vec3(7, 7) == (-5, -4)


def test_sub_scalar_type_error(vcls):
    with pytest.raises(TypeError):
        vcls(1, 1) - 1


def test_isub_scalar_type_erorr(vcls):
    v = vcls(2, 3)
    with pytest.raises(TypeError):
        v -= 1


def test_rsub_tuple(vec2):
    with pytest.raises(TypeError):
        (2, 3) - vec2(7, 7)


def test_rsub_scalar_type_error(vcls):
    with pytest.raises(TypeError):
        1 - vcls(1, 1)


def test_mul_scalar(vcls):
    v = vcls(2, 3)
    assert v * 2 == (4, 6)


def test_imul_scalar(vcls):
    v = vcls(2, 3)
    v *= 2
    assert v == (4, 6)


def test_rmul_scalar(vcls):
    assert 2 * vcls(2, 3) == (4, 6)


def test_mul_tuple_type_error(vcls):
    with pytest.raises(TypeError):
        vcls(2, 3) * (2, 2)


def test_rmul_tuple_type_error(vcls):
    with pytest.raises(TypeError):
        (2, 2) * vcls(2, 3)


def test_imul_tuple_type_error(vcls):
    v = vcls(2, 3)
    with pytest.raises(TypeError):
        v *= (2, 2)


def test_div_scalar(vcls):
    v = vcls(2, 3)
    assert v / 2 == (1, 1.5)


def test_idiv_scalar(vcls):
    v = vcls(2, 3)
    v /= 2
    assert v == (1, 1.5)


def test_dot_product(vcls):
    v1 = vcls(2, 7)
    v2 = vcls(3, 9)
    assert math.isclose(v1.dot(v2), 69)


def test_angle_deg(vcls):
    assert math.isclose(vcls((0, 1)).angle_deg, 90)
    assert math.isclose(vcls((0, -1)).angle_deg, -90)
    assert math.isclose(vcls((1, 1)).angle_deg, 45)
    assert math.isclose(vcls((-1, 1)).angle_deg, 135)


def test_angle_between(vcls):
    v1 = vcls(0, 1)
    v2 = vcls(1, 1)
    angle = v1.angle_between(v2)
    assert math.isclose(angle, math.pi / 4)
    # reverse order, same result
    angle = v2.angle_between(v1)
    assert math.isclose(angle, math.pi / 4)


@pytest.mark.parametrize(
    "v1, v2",
    [
        [(1, 0), (0, 0)],
        [(0, 0), (1, 0)],
        [(0, 0), (0, 0)],
    ],
)
def test_angle_between_null_vector(vcls, v1, v2):
    with pytest.raises(ZeroDivisionError):
        vcls(v1).angle_between(vcls(v2))


def test_angle_between_outside_domain():
    v1 = Vec3(721.046967113573, 721.0469671135688, 0.0)
    v2 = Vec3(-721.0469671135725, -721.0469671135688, 0.0)
    angle = v1.angle_between(v2)
    assert math.isclose(angle, math.pi)
    # reverse order, same result
    angle = v2.angle_between(v1)
    assert math.isclose(angle, math.pi)


def test_rotate(vcls):
    assert vcls(2, 2).rotate_deg(90).isclose(vcls(-2, 2))


def test_lerp(vcls):
    v1 = vcls(1, 1)
    v2 = vcls(4, 4)
    assert v1.lerp(v2, 0.5) == (2.5, 2.5)
    assert v1.lerp(v2, 0) == (1, 1)
    assert v1.lerp(v2, 1) == (4, 4)


def test_project(vcls):
    v = vcls(10, 0)
    assert v.project(vcls(5, 0)) == (5, 0)
    assert v.project(vcls(5, 5)) == (5, 0)
    assert v.project(vcls(5, 5)) == (5, 0)

    v = vcls(10, 10)
    assert v.project(vcls(10, 0)).isclose(vcls(5, 5))


def test_det(vec2):
    assert vec2(1, 0).det(vec2(0, 1)) == 1
    assert vec2(0, 1).det(vec2(1, 0)) == -1


def test_sum(vcls):
    assert vcls.sum([]).is_null is True
    assert vcls.sum([vcls(1, 1)]) == (1, 1)
    assert vcls.sum([vcls(1, 1), vcls(2, 2)]) == (3, 3)


def test_picklable(vec2):
    for v in [vec2((1, 2.5)), vec2(1, 2.5)]:
        pickled_v = pickle.loads(pickle.dumps(v))
        assert v == pickled_v
        assert type(v) is type(pickled_v)



# cython: language_level=3
# Copyright (c) 2021-2024, Manfred Moitzi
# License: MIT License
from typing import Iterable, Sequence, Iterator
import cython
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from .vector cimport Vec3, isclose, v3_mul, v3_sub

__all__ = ['Basis', 'Evaluator']

cdef extern from "constants.h":
    const double ABS_TOL
    const double REL_TOL
    const int MAX_SPLINE_ORDER

# factorial from 0 to 18
cdef double[19] FACTORIAL = [
    1., 1., 2., 6., 24., 120., 720., 5040., 40320., 362880., 3628800.,
    39916800., 479001600., 6227020800., 87178291200., 1307674368000.,
    20922789888000., 355687428096000., 6402373705728000.
]

NULL_LIST = [0.0]
ONE_LIST = [1.0]

cdef Vec3 NULLVEC = Vec3()

@cython.cdivision(True)
cdef double binomial_coefficient(int k, int i):
    cdef double k_fact = FACTORIAL[k]
    cdef double i_fact = FACTORIAL[i]
    cdef double k_i_fact
    if i > k:
        return 0.0
    k_i_fact = FACTORIAL[k - i]
    return k_fact / (k_i_fact * i_fact)

@cython.boundscheck(False)
cdef int bisect_right(double *a, double x, int lo, int hi):
    cdef int mid
    while lo < hi:
        mid = (lo + hi) // 2
        if x < a[mid]:
            hi = mid
        else:
            lo = mid + 1
    return lo

cdef reset_double_array(double *a, int count, double value):
    cdef int i
    for i in range(count):
        a[i] = value

cdef class Basis:
    """ Immutable Basis function class. """
    # public:
    cdef readonly int order
    cdef readonly int count
    cdef readonly double max_t
    cdef tuple weights_  # public attribute for Cython Evaluator
    # private:
    cdef double *_knots
    cdef int knot_count

    def __cinit__(
            self, knots: Iterable[float], 
            int order, 
            int count,
            weights: Sequence[float] = None
        ):
        if order < 2 or order >= MAX_SPLINE_ORDER:
            raise ValueError('invalid order')
        self.order = order
        if count < 2:
            raise ValueError('invalid count')
        self.count = count
        self.knot_count = self.order + self.count
        self.weights_ = tuple(float(x) for x in weights) if weights else tuple()

        cdef Py_ssize_t i = len(self.weights_)
        if i != 0 and i != self.count:
            raise ValueError('invalid weight count')

        knots = [float(x) for x in knots]
        if len(knots) != self.knot_count:
            raise ValueError('invalid knot count')

        self._knots = <double *> PyMem_Malloc(self.knot_count * sizeof(double))
        for i in range(self.knot_count):
            self._knots[i] = knots[i]
        self.max_t = self._knots[self.knot_count - 1]

    def __dealloc__(self):
        PyMem_Free(self._knots)

    @property
    def degree(self) -> int:
        return self.order - 1

    @property
    def knots(self) -> tuple[float, ...]:
        return tuple(x for x in self._knots[:self.knot_count])

    @property
    def weights(self) -> tuple[float, ...]:
        return self.weights_

    @property
    def is_rational(self) -> bool:
        """ Returns ``True`` if curve is a rational B-spline. (has weights) """
        return bool(self.weights_)

    cpdef list basis_vector(self, double t):
        """ Returns the expanded basis vector. """

        cdef int span = self.find_span(t)
        cdef int p = self.order - 1
        cdef int front = span - p
        cdef int back = self.count - span - 1
        cdef list result
        if front > 0:
            result = NULL_LIST * front
            result.extend(self.basis_funcs(span, t))
        else:
            result = self.basis_funcs(span, t)
        if back > 0:
            result.extend(NULL_LIST * back)
        return result

    cpdef int find_span(self, double u):
        """ Determine the knot span index. """
        # Linear search is more reliable than binary search of the Algorithm A2.1
        # from The NURBS Book by Piegl & Tiller.
        cdef double *knots = self._knots
        cdef int count = self.count  # text book: n+1
        cdef int p = self.order - 1
        cdef int span
        if u >= knots[count]:  # special case
            return count - 1

        # common clamped spline:
        if knots[p] == 0.0:  # use binary search
            # This is fast and works most of the time,
            # but Test 621 : test_weired_closed_spline()
            # goes into an infinity loop, because of
            # a weird knot configuration.
            return bisect_right(knots, u, p, count) - 1
        else:  # use linear search
            span = 0
            while knots[span] <= u and span < count:
                span += 1
            return span - 1

    cpdef list basis_funcs(self, int span, double u):
        # Source: The NURBS Book: Algorithm A2.2
        cdef int order = self.order
        cdef double *knots = self._knots
        cdef double[MAX_SPLINE_ORDER] N, left, right  # pyright: ignore
        cdef list result
        reset_double_array(N, order, 0.0)
        reset_double_array(left, order, 0.0)
        reset_double_array(right, order, 0.0)

        cdef int j, r, i1
        cdef double temp, saved, temp_r, temp_l
        N[0] = 1.0
        for j in range(1, order):
            i1 = span + 1 - j
            if i1 < 0:
                i1 = 0
            left[j] = u - knots[i1]
            right[j] = knots[span + j] - u
            saved = 0.0
            for r in range(j):
                temp_r = right[r + 1]
                temp_l = left[j - r]
                temp = N[r] / (temp_r + temp_l)
                N[r] = saved + temp_r * temp
                saved = temp_l * temp
            N[j] = saved
        result = [x for x in N[:order]]
        if self.is_rational:
            return self.span_weighting(result, span)
        else:
            return result

    cpdef list span_weighting(self, nbasis: list[float], int span):
        cdef list products = [
            nb * w for nb, w in zip(
                nbasis,
                self.weights_[span - self.order + 1: span + 1]
            )
        ]
        if len(products) != len(nbasis):
            return nbasis
        s = sum(products)
        if s != 0:
            return [p / s for p in products]
        else:
            return nbasis

    cpdef list basis_funcs_derivatives(self, int span, double u, int n = 1):
        # pyright: reportUndefinedVariable=false
        # pyright flags Cython multi-arrays incorrect: 
        # cdef double[4][4] a  # this is a valid array definition in Cython!
        # https://cython.readthedocs.io/en/latest/src/userguide/language_basics.html#c-arrays
        # Source: The NURBS Book: Algorithm A2.3
        cdef int order = self.order
        cdef int p = order - 1
        if n > p:
            n = p
        cdef double *knots = self._knots
        cdef double[MAX_SPLINE_ORDER] left, right  # pyright: ignore
        reset_double_array(left, order, 1.0)
        reset_double_array(right, order, 1.0)

        cdef double[MAX_SPLINE_ORDER][MAX_SPLINE_ORDER] ndu  # pyright: ignore
        reset_double_array(<double *> ndu, MAX_SPLINE_ORDER*MAX_SPLINE_ORDER, 1.0)

        cdef int j, r, i1
        cdef double temp, saved, tmp_r, tmp_l
        for j in range(1, order):
            i1 = span + 1 - j
            if i1 < 0:
                i1 = 0
            left[j] = u - knots[i1]
            right[j] = knots[span + j] - u
            saved = 0.0
            for r in range(j):
                # lower triangle
                tmp_r = right[r + 1]
                tmp_l = left[j - r]
                ndu[j][r] = tmp_r + tmp_l
                temp = ndu[r][j - 1] / ndu[j][r]
                # upper triangle
                ndu[r][j] = saved + (tmp_r * temp)
                saved = tmp_l * temp
            ndu[j][j] = saved

        # load the basis_vector functions
        cdef double[MAX_SPLINE_ORDER][MAX_SPLINE_ORDER] derivatives  # pyright: ignore
        reset_double_array(
            <double *> derivatives, MAX_SPLINE_ORDER*MAX_SPLINE_ORDER, 0.0
        )
        for j in range(order):
            derivatives[0][j] = ndu[j][p]

        # loop over function index
        cdef double[2][MAX_SPLINE_ORDER] a  # pyright: ignore
        reset_double_array(<double *> a, 2*MAX_SPLINE_ORDER, 1.0)

        cdef int s1, s2, k, rk, pk, j1, j2, t
        cdef double d
        for r in range(order):
            s1 = 0
            s2 = 1
            # alternate rows in array a
            a[0][0] = 1.0

            # loop to compute kth derivative
            for k in range(1, n + 1):
                d = 0.0
                rk = r - k
                pk = p - k
                if r >= k:
                    a[s2][0] = a[s1][0] / ndu[pk + 1][rk]
                    d = a[s2][0] * ndu[rk][pk]
                if rk >= -1:
                    j1 = 1
                else:
                    j1 = -rk
                if (r - 1) <= pk:
                    j2 = k - 1
                else:
                    j2 = p - r
                for j in range(j1, j2 + 1):
                    a[s2][j] = (a[s1][j] - a[s1][j - 1]) / ndu[pk + 1][rk + j]
                    d += (a[s2][j] * ndu[rk + j][pk])
                if r <= pk:
                    a[s2][k] = -a[s1][k - 1] / ndu[pk + 1][r]
                    d += (a[s2][k] * ndu[r][pk])
                derivatives[k][r] = d

                # Switch rows
                t = s1
                s1 = s2
                s2 = t

        # Multiply through by the correct factors
        cdef double rr = p
        for k in range(1, n + 1):
            for j in range(order):
                derivatives[k][j] *= rr
            rr *= (p - k)

        # return result as Python lists
        cdef list result = [], row
        for k in range(0, n + 1):
            row = []
            result.append(row)
            for j in range(order):
                row.append(derivatives[k][j])
        return result

cdef class Evaluator:
    """ B-spline curve point and curve derivative evaluator. """
    cdef Basis _basis
    cdef tuple _control_points

    def __cinit__(self, basis: Basis, control_points: Sequence[Vec3]):
        self._basis = basis
        self._control_points = Vec3.tuple(control_points)

    cpdef Vec3 point(self, double u):
        # Source: The NURBS Book: Algorithm A3.1
        cdef Basis basis = self._basis
        if isclose(u, basis.max_t, REL_TOL, ABS_TOL):
            u = basis.max_t
        cdef:
            int p = basis.order - 1
            int span = basis.find_span(u)
            list N = basis.basis_funcs(span, u)
            int i
            Vec3 cpoint, v3_sum = Vec3()
            tuple control_points = self._control_points
            double factor
            
        for i in range(p + 1):
            factor = <double> N[i]
            cpoint = <Vec3> control_points[span - p + i]
            v3_sum.x += cpoint.x * factor
            v3_sum.y += cpoint.y * factor
            v3_sum.z += cpoint.z * factor
        return v3_sum

    def points(self, t: Iterable[float]) -> Iterator[Vec3]:
        cdef double u
        for u in t:
            yield self.point(u)

    cpdef list derivative(self, double u, int n = 1):
        """ Return point and derivatives up to n <= degree for parameter u. """
        # Source: The NURBS Book: Algorithm A3.2
        
        cdef Basis basis = self._basis
        if isclose(u, basis.max_t, REL_TOL, ABS_TOL):
            u = basis.max_t
        cdef:
            list CK = [], CKw = [], wders = []
            tuple control_points = self._control_points
            tuple weights
            Vec3 cpoint, v3_sum
            double wder, bas_func_weight, bas_func
            int k, j, i, p = basis.degree
            int span = basis.find_span(u)
            list basis_funcs_ders = basis.basis_funcs_derivatives(span, u, n)

        if basis.is_rational:
            # Homogeneous point representation required:
            # (x*w, y*w, z*w, w)
            weights = basis.weights_
            for k in range(n + 1):
                v3_sum = Vec3()
                wder = 0.0
                for j in range(p + 1):
                    i = span - p + j
                    bas_func_weight = basis_funcs_ders[k][j] * weights[i]
                    # control_point * weight * bas_func_der = (x*w, y*w, z*w) * bas_func_der
                    cpoint = <Vec3> control_points[i]
                    v3_sum.x += cpoint.x * bas_func_weight
                    v3_sum.y += cpoint.y * bas_func_weight
                    v3_sum.z += cpoint.z * bas_func_weight
                    wder += bas_func_weight
                CKw.append(v3_sum)
                wders.append(wder)

            # Source: The NURBS Book: Algorithm A4.2
            for k in range(n + 1):
                v3_sum = CKw[k]
                for j in range(1, k + 1):
                    bas_func_weight = binomial_coefficient(k, j) * wders[j]
                    v3_sum = v3_sub(
                        v3_sum,
                        v3_mul(CK[k - j], bas_func_weight)
                    )
                CK.append(v3_sum / wders[0])
        else:
            for k in range(n + 1):
                v3_sum = Vec3()
                for j in range(p + 1):
                    bas_func = basis_funcs_ders[k][j]
                    cpoint = <Vec3> control_points[span - p + j] 
                    v3_sum.x += cpoint.x * bas_func
                    v3_sum.y += cpoint.y * bas_func
                    v3_sum.z += cpoint.z * bas_func
                CK.append(v3_sum)
        return CK

    def derivatives(self, t: Iterable[float], int n = 1) -> Iterator[list[Vec3]]:
        cdef double u
        for u in t:
            yield self.derivative(u, n)



# Copyright (c) 2018-2022 Manfred Moitzi
# License: MIT License
import pytest

import math
from ezdxf.math import Vec3, close_vectors, UCS, BoundingBox, Vec2
from ezdxf.render import forms


def test_circle_open():
    c = list(forms.circle(8))
    assert len(c) == 8


def test_circle_closed():
    c = list(forms.circle(8, close=True))
    assert len(c) == 9


def test_close_polygon():
    p = forms.close_polygon(Vec3.generate([(1, 0), (2, 0), (3, 0), (4, 0)]))
    assert len(p) == 5
    assert p[4] == (1, 0)


def test_close_polygon_without_doublets():
    p = forms.close_polygon(Vec3.generate([(1, 0), (2, 0), (3, 0), (4, 0), (1, 0)]))
    assert len(p) == 5


def test_close_circle():
    assert len(list(forms.circle(8, close=True))) == 9
    assert len(forms.close_polygon(forms.circle(8, close=True))) == 9
    assert len(forms.close_polygon(forms.circle(8, close=False))) == 9


def test_square():
    sq = forms.square(2)
    assert len(sq) == 4
    assert close_vectors(sq, [(0, 0), (2, 0), (2, 2), (0, 2)])


def test_box():
    b = forms.box(3, 2)
    assert len(b) == 4
    assert close_vectors(b, [(0, 0), (3, 0), (3, 2), (0, 2)])


def test_open_arrow():
    a = forms.open_arrow(3, 60)
    assert len(a) == 3
    assert close_vectors(a, [(-3, 1.5), (0, 0), (-3, -1.5)])


def test_closed_arrow():
    a = forms.arrow2(3, 60, 45)
    assert len(a) == 4
    assert close_vectors(a, [(-3, 1.5), (0, 0), (-3, -1.5), (-1.5, 0)])


def test_cube():
    c = forms.cube(center=True)
    assert len(c.vertices) == 8
    assert len(c.faces) == 6

    c = forms.cube(center=False)
    assert len(c.vertices) == 8
    assert len(c.faces) == 6


class TestExtrude:
    @pytest.fixture
    def profile(self):
        return [(0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0)]

    def test_extrude_without_caps(self, profile):
        path = ((0, 0, 0), (0, 0, 1))
        diag = forms.extrude(profile, path, close=True, caps=False).diagnose()
        assert diag.n_vertices == 8
        assert diag.n_faces == 4
        assert diag.is_manifold is True

    def test_extrude_open_profiles_with_caps(self, profile):
        path = ((0, 0, 0), (0, 0, 1))
        diag = forms.extrude(profile, path, close=False, caps=True).diagnose()
        assert diag.n_vertices == 8
        assert diag.n_faces == 5, "hull should have top- and bottom faces"
        assert diag.is_manifold is True

    def test_extrude_with_caps(self, profile):
        path = ((0, 0, 0), (0, 0, 1))
        diag = forms.extrude(profile, path, close=True, caps=True).diagnose()
        assert len(diag.faces[0]) == 4, "bottom face should have 4 vertices"
        assert diag.n_vertices == 8
        assert diag.n_faces == 6
        assert diag.is_manifold is True
        assert len(diag.faces[-1]) == 4, "top face should have 4 vertices"


def test_from_profiles_linear():
    bottom = Vec3.list([(0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0)])
    top = Vec3.list([(0, 0, 1), (1, 0, 1), (1, 1, 1), (0, 1, 1)])
    mesh = forms.from_profiles_linear([bottom, top], close=True, caps=True)
    assert len(mesh.vertices) == 8
    assert len(mesh.faces) == 6

    mesh = forms.from_profiles_linear([bottom, top], close=True, caps=False)
    assert len(mesh.vertices) == 8
    assert len(mesh.faces) == 4


def in_vertices(v, vertices):
    v = Vec3(v)
    return any(v.isclose(v2) for v2 in vertices)


def test_cylinder():
    mesh = forms.cylinder(12)
    assert len(mesh.faces) == 14  # 1x bottom, 1x top, 12x side
    assert len(mesh.vertices) == 24  # 12x bottom, 12x top

    mesh = forms.cylinder(
        count=12, radius=3, top_radius=2, top_center=(1, 0, 3), caps=False
    )
    assert len(mesh.faces) == 12
    assert len(mesh.vertices) == 24
    assert in_vertices((3, 0, 3), mesh.vertices)
    assert in_vertices((-1, 0, 3), mesh.vertices)


def test_spline_interpolation():
    vertices = [(0.0, 0.0), (1.0, 2.0), (3.0, 1.0), (5.0, 3.0)]
    result = forms.spline_interpolation(vertices, method="uniform", subdivide=4)
    assert len(result) == 13  # (len-1) * subdivide + 1
    assert Vec3(0, 0, 0).isclose(result[0]), "expected start point"
    assert Vec3(5, 3, 0).isclose(result[-1]), "expected end point"
    assert Vec3(1, 2, 0).isclose(result[4]), "expected 2. fit point"
    assert Vec3(3, 1, 0).isclose(result[8]), "expected 3. fit point"


def test_spline_interpolated_profiles():
    p1 = forms.circle(12, radius=2, elevation=0, close=True)
    p2 = forms.circle(12, radius=3, elevation=2, close=True)
    p3 = forms.circle(12, radius=1, elevation=4, close=True)
    p4 = forms.circle(12, radius=2, elevation=6, close=True)
    profiles = list(
        forms.spline_interpolated_profiles(
            [Vec3.list(p) for p in [p1, p2, p3, p4]], subdivide=4
        )
    )
    assert len(profiles) == 13  # 3*4 + 1


def test_from_profiles_splines():
    p1 = list(forms.circle(12, radius=2, elevation=0, close=True))
    p2 = list(forms.circle(12, radius=3, elevation=2, close=True))
    p3 = list(forms.circle(12, radius=1, elevation=4, close=True))
    p4 = list(forms.circle(12, radius=2, elevation=6, close=True))
    mesh = forms.from_profiles_spline([p1, p2, p3, p4], subdivide=4, caps=True)
    assert len(mesh.vertices) == 156  # 12 (circle) * 13 (profiles)
    assert len(mesh.faces) == 146  # 12 (circle) * 12 + 2


def test_cone():
    mesh = forms.cone(12, 2, apex=(0, 0, 3))
    assert len(mesh.vertices) == 13
    assert len(mesh.faces) == 13


def test_rotation_form():
    profile = [(0, 0.1), (1, 1), (3, 1.5), (5, 3)]  # in xy-plane
    mesh = forms.rotation_form(
        count=16, profile=profile, axis=(1, 0, 0)
    )  # rotation axis is the x-axis
    assert len(mesh.vertices) == 16 * 4
    assert len(mesh.faces) == 16 * 3


def test_translate():
    p = [(1, 2, 3), (4, 5, 6)]
    r = list(forms.translate(p, (3, 2, 1)))
    assert r[0].isclose((4, 4, 4))
    assert r[1].isclose((7, 7, 7))


def test_scale():
    p = [(1, 2, 3), (4, 5, 6)]
    r = list(forms.scale(p, (3, 2, 1)))
    assert r[0].isclose((3, 4, 3))
    assert r[1].isclose((12, 10, 6))


def test_rotate():
    p = [(1, 0, 3), (0, 1, 6)]
    r = list(forms.rotate(p, 90, deg=True))
    assert r[0].isclose((0, 1, 3))
    assert r[1].isclose((-1, 0, 6))


def test_square_by_radius():
    corners = list(forms.ngon(4, radius=1))
    assert len(corners) == 4
    assert corners[0].isclose((1, 0, 0))
    assert corners[1].isclose((0, 1, 0))
    assert corners[2].isclose((-1, 0, 0))
    assert corners[3].isclose((0, -1, 0))


def test_heptagon_by_edge_length():
    corners = list(forms.ngon(7, length=10))
    assert len(corners) == 7
    assert corners[0].isclose((11.523824354812433, 0, 0))
    assert corners[1].isclose((7.184986963636852, 9.009688679024192, 0))
    assert corners[2].isclose((-2.564292158181384, 11.234898018587335, 0))
    assert corners[3].isclose((-10.382606982861683, 5, 0))
    assert corners[4].isclose((-10.382606982861683, -5, 0))
    assert corners[5].isclose((-2.564292158181387, -11.234898018587335, 0))
    assert corners[6].isclose((7.18498696363685, -9.009688679024192, 0))


class TestTorus:
    def test_closed_torus_ngon_faces(self):
        t = forms.torus(major_count=16, minor_count=8)
        diag = t.diagnose()
        assert diag.n_vertices == 16 * 8
        assert diag.n_faces == 16 * 8
        assert diag.is_manifold is True

    def test_open_torus_ngon_faces(self):
        t = forms.torus(major_count=16, minor_count=8, end_angle=math.pi)
        diag = t.diagnose()
        assert diag.n_vertices == 17 * 8
        assert diag.n_faces == 16 * 8 + 2
        assert diag.is_manifold is True

    @pytest.mark.parametrize("r", [2, 1, -2])
    def test_major_radius_is_bigger_than_minor_radius(self, r):
        with pytest.raises(ValueError):
            forms.torus(major_radius=1, minor_radius=r)

    def test_major_radius_is_bigger_than_zero(self):
        with pytest.raises(ValueError):
            forms.torus(major_radius=0)

    def test_minor_radius_is_bigger_than_zero(self):
        with pytest.raises(ValueError):
            forms.torus(minor_radius=0)


def test_intersection_profiles():
    p0 = Vec3.list([(0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0)])
    p1 = Vec3.list([(0, 0, 1), (1, 0, 1), (1, 1, 1), (0, 1, 1)])
    p2 = Vec3.list([(0, 0, 1), (0, 0, 2), (0, 1, 2), (0, 1, 1)])
    p3 = Vec3.list([(-1, 0, 1), (-1, 0, 2), (-1, 1, 2), (-1, 1, 1)])
    profiles = forms._intersection_profiles([p0, p2], [p1, p3])
    assert profiles[0] == p0
    assert (
        close_vectors(profiles[1], [(0, 0, 1), (1, 0, 2), (1, 1, 2), (0, 1, 1)]) is True
    )
    assert profiles[2] == p3


class TestReferenceFrame:
    def test_ref_z_in_x_axis(self):
        ucs = forms.reference_frame_z(Vec3(1, 0, 0))
        assert ucs.uy.isclose((0, 1, 0))
        assert ucs.uz.isclose((1, 0, 0))
        assert ucs.ux.isclose((0, 0, -1))

    @pytest.mark.parametrize("n", [Vec3(0, 0, 1), Vec3(0, 0, -1)])
    def test_ref_z_in_z_axis_raise_exception(self, n):
        with pytest.raises(ZeroDivisionError):
            forms.reference_frame_z(n)

    def test_ref_ext_preserve_x(self):
        frame = UCS()
        ucs = forms.reference_frame_ext(frame)
        assert ucs.ux.isclose((1, 0, 0))
        assert ucs.uy.isclose((0, 1, 0))
        assert ucs.uz.isclose((0, 0, 1))

    def test_ref_ext_preserve_y(self):
        # x-axis of previous reference frame is parallel to the Z_AXIS
        frame = UCS(ux=(0, 0, 1), uy=(0, 1, 0))
        ucs = forms.reference_frame_ext(frame)
        assert ucs.ux.isclose((1, 0, 0))
        assert ucs.uy.isclose((0, 1, 0))
        assert ucs.uz.isclose((0, 0, 1))


class TestExtrude2:
    def test_divide_path_count(self):
        p = [Vec3(), Vec3(3, 0)]
        assert len(forms._divide_path_into_steps(p, max_step_size=3)) == 2
        assert len(forms._divide_path_into_steps(p, max_step_size=2)) == 3
        assert len(forms._divide_path_into_steps(p, max_step_size=1)) == 4
        assert len(forms._divide_path_into_steps(p, max_step_size=0.5)) == 7

    def test_divide_path_vertices(self):
        p = [Vec3(), Vec3(3, 0)]
        assert (
            close_vectors(
                forms._divide_path_into_steps(p, 1.0),
                Vec3.list([(0, 0), (1, 0), (2, 0), (3, 0)]),
            )
            is True
        )

        p = [Vec3(), Vec3(1, 0), Vec3(2, 0), Vec3(3, 0)]
        assert (
            close_vectors(
                forms._divide_path_into_steps(p, 1.0),
                Vec3.list([(0, 0), (1, 0), (2, 0), (3, 0)]),
            )
            is True
        )

    def test_divide_path_into_equally_spaced_segments(self):
        p = [Vec3(), Vec3(1.5, 0), Vec3(3, 0)]
        assert (
            close_vectors(
                forms._divide_path_into_steps(p, 1.0),
                Vec3.list([(0, 0), (0.75, 0), (1.5, 0), (2.25, 0), (3, 0)]),
            )
            is True
        )

    def test_divide_path_has_to_include_all_source_vertices(self):
        p = [Vec3(), Vec3(1, 0), Vec3(1.5, 0), Vec3(2, 0), Vec3(3, 0)]
        assert (
            close_vectors(
                forms._divide_path_into_steps(p, 1.0),
                Vec3.list([(0, 0), (1, 0), (1.5, 0), (2, 0), (3, 0)]),
            )
            is True
        )

    def test_partial_path_factors(self):
        p = forms._partial_path_factors(
            Vec3.list([(0, 0), (1, 0), (2, 0), (3, 0), (4, 0)])
        )
        assert p == pytest.approx([0, 0.25, 0.5, 0.75, 1.0])

    def test_extrude_square_without_intermediate_profiles(self):
        profile = forms.square(center=True)
        path = [(0, 0), (10, 0)]
        mesh = forms.extrude_twist_scale(profile, path, close=True, step_size=0.0)
        assert len(mesh.vertices) == 8

    def test_extrude_square_with_intermediate_profiles(self):
        profile = forms.square(center=True)
        path = [(0, 0), (0, 0, 10)]
        mesh = forms.extrude_twist_scale(profile, path, close=True, step_size=1.0)
        assert len(mesh.vertices) == 11 * 4
        assert len(mesh.faces) == 10 * 4

    def test_extrude_with_twist(self):
        mesh = forms.extrude_twist_scale(
            profile=forms.square(center=True),
            path=[(0, 0), (0, 0, 10)],
            close=True,
            step_size=1.0,
            twist=math.pi / 2,
        )
        assert close_vectors(
            mesh.vertices[-4:],
            [
                Vec3(0.5, 0.5, 10.0),
                Vec3(0.5, -0.5, 10.0),
                Vec3(-0.5, 0.5, 10.0),
                Vec3(-0.5, -0.5, 10.0),
            ],
        )

    def test_extrude_with_scale(self):
        mesh = forms.extrude_twist_scale(
            profile=forms.square(center=True),
            path=[(0, 0), (0, 0, 10)],
            close=True,
            step_size=1.0,
            scale=2.0,
        )
        assert close_vectors(
            mesh.vertices[-4:],
            [
                Vec3(1.0, -1.0, 10.0),
                Vec3(-1.0, -1.0, 10.0),
                Vec3(1.0, 1.0, 10.0),
                Vec3(-1.0, 1.0, 10.0),
            ],
        )


@pytest.mark.parametrize("ccw", [True, False])
def test_helix_positive_pitch_goes_up(ccw):
    bbox = BoundingBox(forms.helix(radius=2.0, pitch=1.0, turns=10, ccw=ccw))
    assert bbox.extmax.isclose((2, 2, 10))
    assert bbox.extmin.isclose((-2, -2, 0))


@pytest.mark.parametrize("ccw", [True, False])
def test_helix_negative_pitch_goes_down(ccw):
    bbox = BoundingBox(forms.helix(radius=2.0, pitch=-1.0, turns=10, ccw=ccw))
    assert bbox.extmax.isclose((2, 2, 0))
    assert bbox.extmin.isclose((-2, -2, -10))


def test_turtle_turn_left():
    vertices = list(forms.turtle("10 l 10 l 10"))
    assert len(vertices) == 4
    assert vertices[0] == Vec2(0, 0)
    assert vertices[-1].isclose((0, 10))


def test_turtle_turn_right():
    vertices = list(forms.turtle("10 r 10 r 10"))
    assert len(vertices) == 4
    assert vertices[0] == Vec2(0, 0)
    assert vertices[-1].isclose((0, -10))


def test_turtle_move_relative():
    vertices = list(forms.turtle("@10,10"))
    assert len(vertices) == 2
    assert vertices[-1].isclose((10, 10))


class TestCylinder2p:
    def test_default_arguments(self):
        cylinder = forms.cylinder_2p(16)
        diag = cylinder.diagnose()
        assert diag.n_faces == 16 + 2

    def test_cylinder_height_of_0_raises_value_error(self):
        with pytest.raises(ValueError):
            forms.cylinder_2p(16, base_center=(0, 0, 0), top_center=(0, 0, 0))

    @pytest.mark.parametrize(
        "s,e",
        [
            [(1, 2, 3), (3, 2, 1)],
            [(0, 0, -3), (0, 0, 3)],
            [(0, 0, 3), (0, 0, -3)],
            [(1, 0, 0), (3, 0, 0)],
            [(-1, 0, 0), (-3, 0, 0)],
            [(0, 1, 0), (0, 3, 0)],
            [(0, -1, 0), (0, -3, 0)],
        ],
        ids=[
            "arbitrary",
            "+z axis",
            "-z axis",
            "+x axis",
            "-x axis",
            "+y axis",
            "-y axis",
        ],
    )
    def test_various_directions(self, s, e):
        cylinder = forms.cylinder_2p(16, base_center=s, top_center=e)
        diag = cylinder.diagnose()
        assert diag.n_faces == 16 + 2

    def test_final_location_for_negative_z_axis(self):
        cylinder = forms.cylinder_2p(16, base_center=(0, 0, -2), top_center=(0, 0, -4))
        bbox = cylinder.bbox()
        assert bbox.extmax.isclose((1, 1, -2))
        assert bbox.extmin.isclose((-1, -1, -4))


class TestCone2p:
    def test_default_arguments(self):
        cone = forms.cone_2p(16)
        diag = cone.diagnose()
        assert diag.n_faces == 16 + 1

    def test_cylinder_height_of_0_raises_value_error(self):
        with pytest.raises(ValueError):
            forms.cone_2p(16, base_center=(0, 0, 0), apex=(0, 0, 0))

    @pytest.mark.parametrize(
        "s,e",
        [
            [(1, 2, 3), (3, 2, 1)],
            [(0, 0, -3), (0, 0, 3)],
            [(0, 0, 3), (0, 0, -3)],
            [(1, 0, 0), (3, 0, 0)],
            [(-1, 0, 0), (-3, 0, 0)],
            [(0, 1, 0), (0, 3, 0)],
            [(0, -1, 0), (0, -3, 0)],
        ],
        ids=[
            "arbitrary",
            "+z axis",
            "-z axis",
            "+x axis",
            "-x axis",
            "+y axis",
            "-y axis",
        ],
    )
    def test_various_directions(self, s, e):
        cone = forms.cone_2p(16, base_center=s, apex=e)
        diag = cone.diagnose()
        assert diag.n_faces == 16 + 1

    def test_final_location_for_negative_z_axis(self):
        cone = forms.cone_2p(16, base_center=(0, 0, -2), apex=(0, 0, -4))
        bbox = cone.bbox()
        assert bbox.extmax.isclose((1, 1, -2))
        assert bbox.extmin.isclose((-1, -1, -4))



#  Copyright (c) 2021-2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Any, Optional, Iterator, TYPE_CHECKING

from ezdxf import colors
from ezdxf.lldxf import validator, const

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.entities import DXFEntity

__all__ = ["GfxAttribs", "TRANSPARENCY_BYBLOCK"]


DEFAULT_LAYER = "0"
DEFAULT_ACI_COLOR = colors.BYLAYER
DEFAULT_LINETYPE = "ByLayer"
DEFAULT_LINEWEIGHT = const.LINEWEIGHT_BYLAYER
DEFAULT_LTSCALE = 1.0
TRANSPARENCY_BYBLOCK = -1.0  # special value


class GfxAttribs:
    """
    Represents often used DXF attributes of graphical entities.

    Args:
        layer (str): layer name as string
        color (int): :ref:`ACI` color value as integer
        rgb: RGB true color (red, green, blue) tuple, each channel value in the
            range from 0 to 255, ``None`` for not set
        linetype (str): linetype name, does not check if the linetype exist!
        lineweight (int):  see :ref:`lineweights` documentation for valid values
        transparency (float): transparency value in the range from 0.0 to 1.0,
            where 0.0 is opaque and 1.0 if fully transparent, -1.0 for
            transparency by block, ``None`` for transparency by layer
        ltscale (float): linetype scaling factor > 0.0, default factor is 1.0

    Raises:
        DXFValueError: invalid attribute value

    """

    _layer: str = DEFAULT_LAYER
    _aci_color: int = DEFAULT_ACI_COLOR
    _true_color: Optional[colors.RGB] = None
    _linetype: str = DEFAULT_LINETYPE
    _lineweight: int = DEFAULT_LINEWEIGHT
    _transparency: Optional[float] = None
    _ltscale: float = DEFAULT_LTSCALE

    def __init__(
        self,
        *,
        layer: str = DEFAULT_LAYER,
        color: int = DEFAULT_ACI_COLOR,
        rgb: Optional[colors.RGB] = None,
        linetype: str = DEFAULT_LINETYPE,
        lineweight: int = DEFAULT_LINEWEIGHT,
        transparency: Optional[float] = None,
        ltscale: float = DEFAULT_LTSCALE,
    ):
        self.layer = layer
        self.color = color
        self.rgb = rgb
        self.linetype = linetype
        self.lineweight = lineweight
        self.transparency = transparency
        self.ltscale = ltscale

    def __str__(self) -> str:
        s = []
        if self._layer != DEFAULT_LAYER:
            s.append(f"layer='{self._layer}'")
        if self._aci_color != DEFAULT_ACI_COLOR:
            s.append(f"color={self._aci_color}")
        if self._true_color is not None:
            s.append(f"rgb={self._true_color}")
        if self._linetype != DEFAULT_LINETYPE:
            s.append(f"linetype='{self._linetype}'")
        if self._lineweight != DEFAULT_LINEWEIGHT:
            s.append(f"lineweight={self._lineweight}")
        if self._transparency is not None:
            s.append(f"transparency={round(self._transparency, 3)}")
        if self._ltscale != DEFAULT_LTSCALE:
            s.append(f"ltscale={round(self._ltscale, 3)}")

        return ", ".join(s)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__str__()})"

    def __iter__(self) -> Iterator[tuple[str, Any]]:
        """Returns iter(self)."""
        return iter(self.items())

    def items(self, default_values=False) -> list[tuple[str, Any]]:
        """Returns the DXF attributes as list of name, value pairs, returns
        also the default values if argument `default_values` is ``True``.
        The true_color and transparency attributes do not have default values,
        the absence of these attributes is the default value.
        """
        data: list[tuple[str, Any]] = []
        if default_values or self._layer != DEFAULT_LAYER:
            data.append(("layer", self._layer))
        if default_values or self._aci_color != DEFAULT_ACI_COLOR:
            data.append(("color", self._aci_color))
        if self._true_color is not None:
            # absence is the default value
            data.append(("true_color", colors.rgb2int(self._true_color)))
        if default_values or self._linetype != DEFAULT_LINETYPE:
            data.append(("linetype", self._linetype))
        if default_values or self._lineweight != DEFAULT_LINEWEIGHT:
            data.append(("lineweight", self.lineweight))
        if self._transparency is not None:
            # absence is the default value
            if self._transparency == TRANSPARENCY_BYBLOCK:
                data.append(("transparency", colors.TRANSPARENCY_BYBLOCK))
            else:
                data.append(
                    (
                        "transparency",
                        colors.float2transparency(self._transparency),
                    )
                )
        if default_values or self._ltscale != DEFAULT_LTSCALE:
            data.append(("ltscale", self._ltscale))
        return data

    def asdict(self, default_values=False) -> dict[str, Any]:
        """Returns the DXF attributes as :class:`dict`, returns also the
        default values if argument `default_values` is ``True``.
        The true_color and transparency attributes do not have default values,
        the absence of these attributes is the default value.
        """
        return dict(self.items(default_values))

    @property
    def layer(self) -> str:
        """layer name"""
        return self._layer

    @layer.setter
    def layer(self, name: str):
        if validator.is_valid_layer_name(name):
            self._layer = name
        else:
            raise const.DXFValueError(f"invalid layer name '{name}'")

    @property
    def color(self) -> int:
        """:ref:`ACI` color value"""
        return self._aci_color

    @color.setter
    def color(self, value: int):
        if validator.is_valid_aci_color(value):
            self._aci_color = value
        else:
            raise const.DXFValueError(f"invalid ACI color value '{value}'")

    @property
    def rgb(self) -> Optional[colors.RGB]:
        """true color value as (red, green, blue) tuple, ``None`` for not set"""
        return self._true_color

    @rgb.setter
    def rgb(self, value: Optional[colors.RGB]):
        if value is None:
            self._true_color = None
        elif validator.is_valid_rgb(value):
            self._true_color = value
        else:
            raise const.DXFValueError(f"invalid true color value '{value}'")

    @property
    def linetype(self) -> str:
        """linetype name"""
        return self._linetype

    @linetype.setter
    def linetype(self, name: str):
        if validator.is_valid_table_name(name):
            self._linetype = name
        else:
            raise const.DXFValueError(f"invalid linetype name '{name}'")

    @property
    def lineweight(self) -> int:
        """lineweight"""
        return self._lineweight

    @lineweight.setter
    def lineweight(self, value: int):
        if validator.is_valid_lineweight(value):
            self._lineweight = value
        else:
            raise const.DXFValueError(f"invalid lineweight value '{value}'")

    @property
    def transparency(self) -> Optional[float]:
        """transparency value from 0.0 for opaque to 1.0 is fully transparent,
        -1.0 is for transparency by block and ``None`` if for transparency
        by layer
        """
        return self._transparency

    @transparency.setter
    def transparency(self, value: Optional[float]):
        if value is None:
            self._transparency = None
        elif value == TRANSPARENCY_BYBLOCK:
            self._transparency = TRANSPARENCY_BYBLOCK
        elif isinstance(value, float) and (0.0 <= value <= 1.0):
            self._transparency = value
        else:
            raise const.DXFValueError(f"invalid transparency value '{value}'")

    @property
    def ltscale(self) -> float:
        """linetype scaling factor"""
        return self._ltscale

    @ltscale.setter
    def ltscale(self, value: float):
        if isinstance(value, (float, int)) and (value > 1e-6):
            self._ltscale = float(value)
        else:
            raise const.DXFValueError(f"invalid linetype scale value '{value}'")

    @classmethod
    def load_from_header(cls, doc: Drawing) -> GfxAttribs:
        """Load default DXF attributes from the HEADER section.

        There is no default true color value and the default transparency is not
        stored in the HEADER section.

        Loads following header variables:

            - ``$CLAYER`` - current layer name
            - ``$CECOLOR`` - current ACI color
            - ``$CELTYPE`` - current linetype name
            - ``$CELWEIGHT`` - current lineweight
            - ``$CELTSCALE`` - current linetype scaling factor

        """
        header = doc.header
        return cls(
            layer=header.get("$CLAYER", DEFAULT_LAYER),
            color=header.get("$CECOLOR", DEFAULT_ACI_COLOR),
            linetype=header.get("$CELTYPE", DEFAULT_LINETYPE),
            lineweight=header.get("$CELWEIGHT", DEFAULT_LINEWEIGHT),
            ltscale=header.get("$CELTSCALE", DEFAULT_LTSCALE),
        )

    def write_to_header(self, doc: "Drawing") -> None:
        """Write DXF attributes as default values to the HEADER section.

        Writes following header variables:

            - ``$CLAYER`` - current layer name, if a layer table entry exist in `doc`
            - ``$CECOLOR`` - current ACI color
            - ``$CELTYPE`` - current linetype name, if a linetype table entry exist in `doc`
            - ``$CELWEIGHT`` - current lineweight
            - ``$CELTSCALE`` - current linetype scaling factor

        """
        header = doc.header
        if doc.layers.has_entry(self.layer):
            header["$CLAYER"] = self.layer
        header["$CECOLOR"] = self.color
        if doc.linetypes.has_entry(self.linetype):
            header["$CELTYPE"] = self.linetype
        header["$CELWEIGHT"] = self.lineweight
        header["$CELTSCALE"] = self.ltscale

    @classmethod
    def from_entity(cls, entity: DXFEntity) -> GfxAttribs:
        """Get the graphical attributes of an `entity` as :class:`GfxAttribs`
        object.
        """
        try:
            d = entity.graphic_properties()  # type: ignore
        except AttributeError:
            return cls()
        return cls.from_dict(d)

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> GfxAttribs:
        """Construct :class:`GfxAttribs` from a dictionary of raw DXF values.

        Supported attributes are:

            - layer: layer name as string
            - color: :ref:`ACI` value as int
            - true_color: raw DXF integer value for RGB colors
            - rgb: RGB tuple of int or ``None``
            - linetype: linetype name as string
            - lineweight: lineweight as int, see basic concept of :ref:`lineweights`
            - transparency: raw DXF integer value of transparency or a float in the
              range from 0.0 to 1.0
            - ltscale: linetype scaling factor as float

        """
        attribs = cls()
        for attrib_name in [
            "layer",
            "color",
            "linetype",
            "lineweight",
            "ltscale",
            "rgb",
        ]:
            if attrib_name in d:
                setattr(attribs, attrib_name, d[attrib_name])
        if "true_color" in d:
            attribs.rgb = colors.int2rgb(d["true_color"])
        if "transparency" in d:
            transparency = d["transparency"]
            if isinstance(transparency, float):
                attribs.transparency = transparency
            elif transparency == colors.TRANSPARENCY_BYBLOCK:
                attribs.transparency = TRANSPARENCY_BYBLOCK
            else:
                attribs.transparency = colors.transparency2float(transparency)
        return attribs



#  Copyright (c) 2020-2022, Manfred Moitzi
#  License: MIT License
import pytest
import math
from ezdxf.layouts import VirtualLayout
from ezdxf.math import Matrix44, OCS, Vec3, close_vectors
from ezdxf.path import (
    Path,
    bbox,
    precise_bbox,
    fit_paths_into_box,
    transform_paths,
    transform_paths_to_ocs,
    to_polylines3d,
    to_lines,
    to_lwpolylines,
    to_polylines2d,
    to_hatches,
    to_mpolygons,
    to_bsplines_and_vertices,
    to_splines_and_polylines,
    from_vertices,
    to_multi_path,
    single_paths,
    lines_to_curve3,
    lines_to_curve4,
    is_rectangular,
)
from ezdxf.path import make_path, Command
from ezdxf.entities import BoundaryPathType, EdgeType
from ezdxf.render import forms


class TestTransformPaths:
    def test_empty_paths(self):
        result = transform_paths([], Matrix44())
        assert len(result) == 0

    def test_start_point_only_paths(self):
        result = transform_paths([Path((1, 2, 3))], Matrix44())
        assert len(result) == 1
        assert len(result[0]) == 0
        assert result[0].start == (1, 2, 3)

    def test_transformation_is_executed(self):
        # Real transformation is just tested once, because Matrix44
        # transformation is tested in 605:
        result = transform_paths([Path((1, 2, 3))], Matrix44.translate(1, 1, 1))
        assert result[0].start == (2, 3, 4)

    def test_one_path_line_to(self):
        path = Path()
        path.line_to((1, 0))
        result = transform_paths([path], Matrix44())
        path0 = result[0]
        assert path0[0].type == Command.LINE_TO
        assert path0.start == (0, 0)
        assert path0.end == (1, 0)

    def test_one_path_curve3_to(self):
        path = Path()
        path.curve3_to((2, 0), (1, 1))
        result = transform_paths([path], Matrix44())
        path0 = result[0]
        assert path0[0].type == Command.CURVE3_TO
        assert len(path0[0]) == 2
        assert path0.start == (0, 0)
        assert path0.end == (2, 0)

    def test_one_path_curve4_to(self):
        path = Path()
        path.curve4_to((2, 0), (0, 1), (2, 1))
        result = transform_paths([path], Matrix44())
        path0 = result[0]
        assert path0[0].type == Command.CURVE4_TO
        assert len(path0[0]) == 3
        assert path0.start == (0, 0)
        assert path0.end == (2, 0)

    def test_one_path_multiple_command(self):
        path = Path()
        path.line_to((1, 0))
        path.curve3_to((2, 0), (2.5, 1))
        path.curve4_to((3, 0), (2, 1), (3, 1))
        result = transform_paths([path], Matrix44())

        path0 = result[0]
        assert path0[0].type == Command.LINE_TO
        assert path0[1].type == Command.CURVE3_TO
        assert path0[2].type == Command.CURVE4_TO
        assert path0.start == (0, 0)
        assert path0.end == (3, 0)

    def test_two_paths_one_command(self):
        path_a = Path()
        path_a.line_to((1, 0))
        path_b = Path((2, 0))
        path_b.line_to((3, 0))
        result = transform_paths([path_a, path_b], Matrix44())

        path0 = result[0]
        assert path0[0].type == Command.LINE_TO
        assert path0.start == (0, 0)
        assert path0.end == (1, 0)

        path1 = result[1]
        assert path1[0].type == Command.LINE_TO
        assert path1.start == (2, 0)
        assert path1.end == (3, 0)

    def test_two_paths_multiple_commands(self):
        path_a = Path()
        path_a.line_to((1, 0))
        path_a.curve3_to((2, 0), (2.5, 1))
        path_a.curve4_to((3, 0), (2, 1), (3, 1))

        path_b = path_a.transform(Matrix44.translate(4, 0, 0))
        result = transform_paths([path_a, path_b], Matrix44())

        path0 = result[0]
        assert path0[0].type == Command.LINE_TO
        assert path0[1].type == Command.CURVE3_TO
        assert path0[2].type == Command.CURVE4_TO
        assert path0.start == (0, 0)
        assert path0.end == (3, 0)

        path1 = result[1]
        assert path1[0].type == Command.LINE_TO
        assert path1[1].type == Command.CURVE3_TO
        assert path1[2].type == Command.CURVE4_TO
        assert path1.start == (4, 0)
        assert path1.end == (7, 0)

    def test_multi_path_objects(self):
        path = Path()
        path.line_to((1, 0, 0))
        path.move_to((2, 0, 0))
        paths = transform_paths([path], Matrix44.translate(0, 1, 0))
        assert len(paths) == 1
        path2 = paths[0]
        assert path2.start.isclose((0, 1, 0))
        assert len(path2) == 2
        assert path2.end.isclose((2, 1, 0))
        assert path2.has_sub_paths is True

    def test_to_ocs(self):
        p = Path((0, 1, 1))
        p.line_to((0, 1, 3))
        ocs = OCS((1, 0, 0))  # x-Axis
        result = list(transform_paths_to_ocs([p], ocs))
        p0 = result[0]
        assert ocs.from_wcs((0, 1, 1)) == p0.start
        assert ocs.from_wcs((0, 1, 3)) == p0[0].end


class TestPreciseBoundingBox:
    def test_empty_path(self):
        result = precise_bbox(Path())
        assert result.has_data is False

    def test_line_to(self):
        p = Path()
        p.line_to((1, 2, 3))
        result = precise_bbox(p)
        assert result.size == (1, 2, 3)

    def test_curve3_to(self):
        p = Path()
        p.curve3_to((2, 0), (1, 1))  # end, ctrl
        result = precise_bbox(p)
        assert result.extmax.y == pytest.approx(0.5)  # parabola

    def test_curve4_to(self):
        p = Path()
        p.curve4_to((2, 0), (0, 1), (2, 1))  # end, ctrl1, ctrl2
        result = precise_bbox(p)
        assert result.extmax.y == pytest.approx(0.75)

    def test_move_to(self):
        p = Path()
        p.line_to((1, 0, 0))
        p.move_to((1, 2, 3))
        p.line_to((1, 0, 0))
        result = precise_bbox(p)
        assert result.size == (1, 2, 3)


class TestBoundingBox:
    def test_empty_paths(self):
        result = bbox([])
        assert result.has_data is False

    def test_one_path(self):
        p = Path()
        p.line_to((1, 2, 3))
        assert bbox([p]).size == (1, 2, 3)

    def test_two_path(self):
        p1 = Path()
        p1.line_to((1, 2, 3))
        p2 = Path()
        p2.line_to((-3, -2, -1))
        assert bbox([p1, p2]).size == (4, 4, 4)

    @pytest.fixture(scope="class")
    def quadratic(self):
        p = Path()
        p.curve3_to((2, 0), (1, 1))
        return p

    def test_not_precise_box(self, quadratic):
        result = bbox([quadratic], fast=True)
        assert result.extmax.y == pytest.approx(1)  # control point

    def test_precise_box(self, quadratic):
        result = bbox([quadratic], fast=False)
        assert result.extmax.y == pytest.approx(0.5)  # parabola


class TestFitPathsIntoBoxUniformScaling:
    @pytest.fixture(scope="class")
    def spath(self):
        p = Path()
        p.line_to((1, 2, 3))
        return p

    def test_empty_paths(self):
        assert fit_paths_into_box([], (0, 0, 0)) == []

    def test_uniform_stretch_paths_limited_by_z(self, spath):
        result = fit_paths_into_box([spath], (6, 6, 6))
        box = bbox(result)
        assert box.size == (2, 4, 6)

    def test_uniform_stretch_paths_limited_by_y(self, spath):
        result = fit_paths_into_box([spath], (6, 3, 6))
        box = bbox(result)
        # stretch factor: 1.5
        assert box.size == (1.5, 3, 4.5)

    def test_uniform_stretch_paths_limited_by_x(self, spath):
        result = fit_paths_into_box([spath], (1.2, 6, 6))
        box = bbox(result)
        # stretch factor: 1.2
        assert box.size.isclose((1.2, 2.4, 3.6))

    def test_uniform_shrink_paths(self, spath):
        result = fit_paths_into_box([spath], (1.5, 1.5, 1.5))
        box = bbox(result)
        assert box.size.isclose((0.5, 1, 1.5))

    def test_project_into_xy(self, spath):
        result = fit_paths_into_box([spath], (6, 6, 0))
        box = bbox(result)
        # Note: z-axis is also ignored by extent detection:
        # scaling factor = 3x
        assert box.size.isclose((3, 6, 0)), "z-axis should be ignored"

    def test_project_into_xz(self, spath):
        result = fit_paths_into_box([spath], (6, 0, 6))
        box = bbox(result)
        assert box.size.isclose((2, 0, 6)), "y-axis should be ignored"

    def test_project_into_yz(self, spath):
        result = fit_paths_into_box([spath], (0, 6, 6))
        box = bbox(result)
        assert box.size.isclose((0, 4, 6)), "x-axis should be ignored"

    def test_invalid_target_size(self, spath):
        with pytest.raises(ValueError):
            fit_paths_into_box([spath], (0, 0, 0))


class TestFitPathsIntoBoxNonUniformScaling:
    @pytest.fixture(scope="class")
    def spath(self):
        p = Path()
        p.line_to((1, 2, 3))
        return p

    def test_non_uniform_stretch_paths(self, spath):
        result = fit_paths_into_box([spath], (8, 7, 6), uniform=False)
        box = bbox(result)
        assert box.size == (8, 7, 6)

    def test_non_uniform_shrink_paths(self, spath):
        result = fit_paths_into_box([spath], (1.5, 1.5, 1.5), uniform=False)
        box = bbox(result)
        assert box.size == (1.5, 1.5, 1.5)

    def test_project_into_xy(self, spath):
        result = fit_paths_into_box([spath], (6, 6, 0), uniform=False)
        box = bbox(result)
        assert box.size == (6, 6, 0), "z-axis should be ignored"

    def test_project_into_xz(self, spath):
        result = fit_paths_into_box([spath], (6, 0, 6), uniform=False)
        box = bbox(result)
        assert box.size == (6, 0, 6), "y-axis should be ignored"

    def test_project_into_yz(self, spath):
        result = fit_paths_into_box([spath], (0, 6, 6), uniform=False)
        box = bbox(result)
        assert box.size == (0, 6, 6), "x-axis should be ignored"


class TestPathToBsplineAndVertices:
    def test_empty_path(self):
        result = list(to_bsplines_and_vertices(Path()))
        assert result == []

    def test_only_vertices(self):
        p = from_vertices([(1, 0), (2, 0), (3, 1)])
        result = list(to_bsplines_and_vertices(p))
        assert len(result) == 1, "expected one list of vertices"
        assert len(result[0]) == 3, "expected 3 vertices"

    def test_one_quadratic_bezier(self):
        p = Path()
        p.curve3_to((4, 0), (2, 2))
        result = list(to_bsplines_and_vertices(p))
        assert len(result) == 1, "expected one B-spline"
        cpnts = result[0].control_points
        # A quadratic bezier should be converted to cubic bezier curve, which
        # has a precise cubic B-spline representation.
        assert len(cpnts) == 4, "expected 4 control vertices"
        assert cpnts[0] == (0, 0)
        assert cpnts[3] == (4, 0)

    def test_one_cubic_bezier(self):
        p = Path()
        p.curve4_to((4, 0), (1, 2), (3, 2))
        result = list(to_bsplines_and_vertices(p))
        assert len(result) == 1, "expected one B-spline"
        # cubic bezier curve maps 1:1 to cubic B-spline curve
        # see tests: 630b for the bezier_to_bspline() function

    def test_adjacent_cubic_beziers_with_G1_continuity(self):
        p = Path()
        p.curve4_to((4, 0), (1, 2), (3, 2))
        p.curve4_to((8, 0), (5, -2), (7, -2))
        result = list(to_bsplines_and_vertices(p))
        assert len(result) == 1, "expected one B-spline"
        # cubic bezier curve maps 1:1 to cubic B-spline curve
        # see tests: 630b for the bezier_to_bspline() function

    def test_adjacent_cubic_beziers_without_G1_continuity(self):
        p = Path()
        p.curve4_to((4, 0), (1, 2), (3, 2))
        p.curve4_to((8, 0), (5, 2), (7, 2))
        result = list(to_bsplines_and_vertices(p))
        assert len(result) == 2, "expected two B-splines"

    def test_multiple_segments(self):
        p = Path()
        p.curve4_to((4, 0), (1, 2), (3, 2))
        p.line_to((6, 0))
        p.curve3_to((8, 0), (7, 1))
        result = list(to_bsplines_and_vertices(p))
        assert len(result) == 3, "expected three segments"


class TestToEntityConverter:
    @pytest.fixture
    def path(self):
        p = Path()
        p.line_to((4, 0, 0))
        p.curve4_to((0, 0, 0), (3, 1, 1), (1, 1, 1))
        return p

    @pytest.fixture
    def path1(self):
        p = Path((0, 0, 1))
        p.curve4_to((4, 0, 1), (1, 1, 1), (3, 1, 1))
        return p

    def test_empty_to_polylines3d(self):
        assert list(to_polylines3d([])) == []

    def test_to_polylines3d(self, path):
        polylines = list(to_polylines3d(path))
        assert len(polylines) == 1
        p0 = polylines[0]
        assert p0.dxftype() == "POLYLINE"
        assert p0.is_3d_polyline is True
        assert len(p0) == 18
        assert p0.vertices[0].dxf.location == (0, 0, 0)
        assert p0.vertices[-1].dxf.location == (0, 0, 0)

    def test_empty_to_lines(self):
        assert list(to_lines([])) == []

    def test_to_lines(self, path):
        lines = list(to_lines(path))
        assert len(lines) == 17
        l0 = lines[0]
        assert l0.dxftype() == "LINE"
        assert l0.dxf.start == (0, 0, 0)
        assert l0.dxf.end == (4, 0, 0)

    def test_empty_to_lwpolyline(self):
        assert list(to_lwpolylines([])) == []

    def test_empty_path_to_lwpolyline(self):
        assert list(to_lwpolylines([Path()])) == []

    def test_to_lwpolylines(self, path):
        polylines = list(to_lwpolylines(path))
        assert len(polylines) == 1
        p0 = polylines[0]
        assert p0.dxftype() == "LWPOLYLINE"
        assert tuple(p0[0]) == (0, 0, 0, 0, 0)  # x, y, swidth, ewidth, bulge
        assert tuple(p0[-1]) == (0, 0, 0, 0, 0)

    def test_to_lwpolylines_with_wcs_elevation(self, path1):
        polylines = list(to_lwpolylines(path1))
        p0 = polylines[0]
        assert p0.dxf.elevation == 1

    def test_to_lwpolylines_with_ocs(self, path1):
        m = Matrix44.x_rotate(math.pi / 4)
        path = path1.transform(m)
        extrusion = m.transform((0, 0, 1))
        polylines = list(to_lwpolylines(path, extrusion=extrusion))
        p0 = polylines[0]
        assert p0.dxf.elevation == pytest.approx(1)
        assert p0.dxf.extrusion.isclose(extrusion)
        assert (
            all(
                math.isclose(a, b, abs_tol=1e-12)
                for a, b in zip(p0[0], (0, 0, 0, 0, 0))
            )
            is True
        )
        assert (
            all(
                math.isclose(a, b, abs_tol=1e-12)
                for a, b in zip(p0[-1], (4, 0, 0, 0, 0))
            )
            is True
        )

    def test_multi_path_to_lwpolylines(self):
        path = Path()
        path.line_to((1, 0, 0))
        path.move_to((2, 0, 0))
        path.line_to((3, 0, 0))
        polylines = list(to_lwpolylines(path))
        assert len(polylines) == 2
        assert len(polylines[0]) == 2
        assert len(polylines[1]) == 2

    def test_empty_to_polylines2d(self):
        assert list(to_polylines2d([])) == []

    def test_to_polylines2d(self, path):
        polylines = list(to_polylines2d(path))
        assert len(polylines) == 1
        p0 = polylines[0]
        assert p0.dxftype() == "POLYLINE"
        assert p0.is_2d_polyline is True
        assert p0[0].dxf.location == (0, 0, 0)
        assert p0[-1].dxf.location == (0, 0, 0)

    def test_to_polylines2d_with_wcs_elevation(self, path1):
        polylines = list(to_polylines2d(path1))
        p0 = polylines[0]
        assert p0.dxf.elevation == (0, 0, 1)

    def test_to_polylines2d_with_ocs(self, path1):
        m = Matrix44.x_rotate(math.pi / 4)
        path = path1.transform(m)
        extrusion = m.transform((0, 0, 1))
        polylines = list(to_polylines2d(path, extrusion=extrusion))
        p0 = polylines[0]
        assert p0.dxf.elevation.isclose((0, 0, 1))
        assert p0.dxf.extrusion.isclose(extrusion)
        assert p0[0].dxf.location.isclose((0, 0, 1))
        assert p0[-1].dxf.location.isclose((4, 0, 1))

    def test_empty_to_hatches(self):
        assert list(to_hatches([])) == []

    def test_to_poly_path_hatches(self, path):
        hatches = list(to_hatches(path, edge_path=False))
        assert len(hatches) == 1
        h0 = hatches[0]
        assert h0.dxftype() == "HATCH"
        assert len(h0.paths) == 1

    def test_to_poly_path_hatches_with_wcs_elevation(self, path1):
        hatches = list(to_hatches(path1, edge_path=False))
        ho = hatches[0]
        assert ho.dxf.elevation.isclose((0, 0, 1))

    def test_to_poly_path_hatches_with_ocs(self, path1):
        m = Matrix44.x_rotate(math.pi / 4)
        path = path1.transform(m)
        extrusion = m.transform((0, 0, 1))
        hatches = list(to_hatches(path, edge_path=False, extrusion=extrusion))
        h0 = hatches[0]
        assert h0.dxf.elevation.isclose((0, 0, 1))
        assert h0.dxf.extrusion.isclose(extrusion)
        polypath0 = h0.paths[0]
        assert (
            all(abs(a) < 1e-12 for a in polypath0.vertices[0])
            is True  # ~(0, 0, 0)
        )  # x, y, bulge
        assert (
            all(abs(a) < 1e-12 for a in polypath0.vertices[-1])
            is True  # ~(0, 0, 0)
        ), "should be closed automatically"

    def test_to_edge_path_hatches(self, path):
        hatches = list(to_hatches(path, edge_path=True))
        assert len(hatches) == 1
        h0 = hatches[0]
        assert h0.dxftype() == "HATCH"
        assert len(h0.paths) == 1
        edge_path = h0.paths[0]
        assert edge_path.type == BoundaryPathType.EDGE
        line, spline = edge_path.edges
        assert line.type == EdgeType.LINE
        assert line.start == (0, 0)
        assert line.end == (4, 0)
        assert spline.type == EdgeType.SPLINE
        assert close_vectors(
            Vec3.generate(spline.control_points),
            [(4, 0), (3, 1), (1, 1), (0, 0)],
        )

    def test_to_splines_and_polylines(self, path):
        entities = list(to_splines_and_polylines([path]))
        assert len(entities) == 2
        polyline = entities[0]
        spline = entities[1]
        assert polyline.dxftype() == "POLYLINE"
        assert spline.dxftype() == "SPLINE"
        assert polyline.vertices[0].dxf.location.isclose((0, 0))
        assert polyline.vertices[1].dxf.location.isclose((4, 0))
        assert close_vectors(
            Vec3.generate(spline.control_points),
            [(4, 0, 0), (3, 1, 1), (1, 1, 1), (0, 0, 0)],
        )

    def test_to_mpolygons_returns_expected_dxf_type(self, path):
        # Works internally like to_hatches() but with polyline paths
        # as boundaries only.
        polygons = list(
            to_mpolygons(
                path,
                dxfattribs={
                    "color": 6,  # boundary line color
                    "fill_color": 1,
                },
            )
        )
        assert len(polygons) == 1
        mp = polygons[0]
        assert mp.dxftype() == "MPOLYGON"
        assert len(mp.paths) == 1
        assert mp.dxf.color == 6
        assert mp.dxf.fill_color == 1


# Issue #224 regression test
@pytest.fixture
def ellipse():
    layout = VirtualLayout()
    return layout.add_ellipse(
        center=(1999.488177113287, -1598.02265357955, 0.0),
        major_axis=(629.968069297, 0.0, 0.0),
        ratio=0.495263197,
        start_param=-1.261396328799999,
        end_param=-0.2505454928,
        dxfattribs={
            "layer": "0",
            "linetype": "Continuous",
            "color": 3,
            "extrusion": (0.0, 0.0, -1.0),
        },
    )


def test_to_multi_path():
    p0 = Path((1, 0, 0))
    p0.line_to((2, 0, 0))
    p0.move_to((3, 0, 0))  # will be replaced by move_to(4, 0, 0)
    p1 = Path((4, 0, 0))
    p1.line_to((5, 0, 0))
    p1.move_to((6, 0, 0))
    path = to_multi_path([p0, p1])
    assert path.has_sub_paths is True
    assert path.start == (1, 0, 0)
    assert path.end == (6, 0, 0)
    assert path[1].type == Command.MOVE_TO
    assert path[1].end == (4, 0, 0)


def test_to_multi_path_ignores_empty_paths():
    p0 = Path((1, 0, 0))
    p0.line_to((2, 0, 0))
    empty = Path((100, 0, 0))
    path = to_multi_path([p0, empty])
    assert len(path) == 1
    assert path.has_sub_paths is False
    assert path.end.isclose((2, 0, 0))


def test_single_paths_from_a_single_path_object():
    p = Path((1, 0, 0))
    assert len(list(single_paths([p]))) == 1


def test_single_paths_from_a_multi_path_object():
    p = Path((1, 0, 0))
    p.line_to((2, 0, 0))  # 1st sub-path
    p.move_to((3, 0, 0))  # 2nd sub-path
    p.line_to((4, 0, 0))
    p.move_to((5, 0, 0))  # 3rd sub-path
    paths = list(single_paths([p]))
    assert len(paths) == 3


def test_issue_224_end_points(ellipse):
    p = make_path(ellipse)

    assert ellipse.start_point.isclose(p.start)
    assert ellipse.end_point.isclose(p.end)

    # end point locations measured in BricsCAD:
    assert ellipse.start_point.isclose((2191.3054, -1300.8375), abs_tol=1e-4)
    assert ellipse.end_point.isclose((2609.7870, -1520.6677), abs_tol=1e-4)


def test_issue_494_make_path_from_spline_defined_by_fit_points_and_tangents():
    from ezdxf.entities import Spline

    spline = Spline.new(
        dxfattribs={
            "degree": 3,
            "start_tangent": (0.9920663924871818, 0.1257150464243202, 0.0),
            "end_tangent": (0.9999448476387669, -0.0105024606965807, 0.0),
        },
    )
    spline.fit_points = [
        (209.5080107190219, 206.963463282597, 0.0),
        (209.55254921431026, 206.96662062623636, 0.0),
    ]
    p = make_path(spline)
    assert len(p) > 0


class TestAllLinesToCurveConverter:
    def test_create_a_curve3_command(self):
        path = Path()
        path.line_to((1, 0))
        path = lines_to_curve3(path)
        assert path[0].type == Command.CURVE3_TO

    def test_create_a_curve4_command(self):
        path = Path()
        path.line_to((1, 0))
        path = lines_to_curve4(path)
        assert path[0].type == Command.CURVE4_TO

    @pytest.mark.parametrize(
        "func",
        [
            lines_to_curve3,
            lines_to_curve4,
        ],
    )
    def test_line_to_curve_creates_a_linear_segment(self, func):
        v1, v2 = 1, 2
        path = Path(start=(v1, v1, v1))
        path.line_to((v2, v2, v2))
        path = func(path)
        vertices = list(path.flattening(1))
        assert len(vertices) > 2
        assert all(
            [
                math.isclose(v.x, v.y) and math.isclose(v.x, v.z)
                for v in vertices
            ]
        ), "all vertices have to be located along a line (x == y == z)"

    def test_remove_line_segments_of_zero_length_at_the_start(self):
        # CURVE3_TO and CURVE4_TO can not process zero length segments
        path = Path()
        path.line_to((0, 0))  # line segment of length==0 should be removed
        path.line_to((1, 0))
        path = lines_to_curve4(path)
        assert len(path) == 1
        assert path.start == (0, 0)
        assert path[0].type == Command.CURVE4_TO
        assert path[0].end == (1, 0)

    def test_remove_line_segments_of_zero_length_between_commands(self):
        # CURVE3_TO and CURVE4_TO can not process zero length segments
        path = Path()
        path.line_to((1, 0))
        path.line_to((1, 0))  # line segment of length==0 should be removed
        path.line_to((2, 0))
        path = lines_to_curve4(path)
        assert len(path) == 2
        assert path.start == (0, 0)
        assert path[0].type == Command.CURVE4_TO
        assert path[0].end == (1, 0)
        assert path[1].type == Command.CURVE4_TO
        assert path[1].end == (2, 0)

    def test_remove_line_segments_of_zero_length_at_the_end(self):
        # CURVE3_TO and CURVE4_TO can not process zero length segments
        path = Path()
        path.line_to((1, 0))
        path.line_to((1, 0))  # line segment of length==0 should be removed
        path = lines_to_curve4(path)
        assert len(path) == 1
        assert path.start == (0, 0)
        assert path[0].type == Command.CURVE4_TO
        assert path[0].end == (1, 0)

    def test_does_not_remove_a_line_representing_a_single_point(self):
        path = Path((1, 0))
        path.line_to((1, 0))  # represents the point (1, 0)
        path = lines_to_curve4(path)
        assert len(path) == 1
        assert path[0].type == Command.LINE_TO

    @pytest.mark.parametrize(
        "start,delta",
        [
            (0, 1e-11),  # uses absolute tolerance of 1e-12 near zero!
            (10, 1e-8),  # uses relative tolerance of 1e-9 away from zero!
        ],
    )
    def test_for_very_short_line_segments(self, start, delta):
        path = Path((start, 0, 0))
        path.line_to((start + delta, 0, 0))
        path = lines_to_curve4(path)
        assert len(path) == 1
        assert path[0].type == Command.CURVE4_TO
        assert len(list(path.flattening(1))) > 3

    @pytest.mark.parametrize(
        "start,delta",
        [
            (0, 1e-12),  # uses absolute tolerance of 1e-12 near zero!
            (10, 1e-9),  # uses relative tolerance of 1e-9 away from zero!
        ],
    )
    def test_which_length_is_too_short_to_create_a_curve(self, start, delta):
        path = Path((start, 0, 0))
        path.line_to((start + delta, 0, 0))
        path = lines_to_curve4(path)
        assert len(path) == 1
        assert (
            path[0].type == Command.LINE_TO
        ), "should not remove a single line segment representing a point"
        assert len(list(path.flattening(1))) == 2


class TestIsRectangular:
    def test_empty_path(self):
        assert is_rectangular(Path()) is False

    def test_less_than_four_corners(self):
        p = from_vertices([(0, 0), (1, 0), (1, 1)])
        assert is_rectangular(p) is False

    def test_open_square(self):
        p = from_vertices([(0, 0), (1, 0), (1, 1), (0, 1)])
        assert is_rectangular(p) is True

    def test_closed_square(self):
        p = from_vertices([(0, 0), (1, 0), (1, 1), (0, 1), (0, 0)])
        assert is_rectangular(p) is True

    def test_rectangle(self):
        p = from_vertices([(0, 0), (2, 0), (2, 1), (0, 1)])
        assert is_rectangular(p) is True

    def test_parallelogram(self):
        p = from_vertices([(0, 0), (2, 0), (3, 1), (1, 1)])
        assert is_rectangular(p) is False

    def test_non_aligned_square(self):
        p = from_vertices(forms.rotate(forms.square(2), 30))
        assert is_rectangular(p, aligned=False) is True


def test_polyline_with_bulge_value_greater_one():
    msp = VirtualLayout()
    pline = msp.add_lwpolyline([(0, 0, 0, 0, 2), (1, 0)])
    p = make_path(pline, segments=12)

    assert len(p) == 4



# Copyright (c) 2014-2023, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.colors import int2rgb, rgb2int, aci2rgb, luminance, RGB, RGBA


def test_rgb():
    r, g, b = int2rgb(0xA0B0C0)
    assert 0xA0 == r
    assert 0xB0 == g
    assert 0xC0 == b


def test_from_rgb():
    assert 0xA0B0C0 == rgb2int((0xA0, 0xB0, 0xC0))


def test_from_aci():
    assert (255, 0, 0) == aci2rgb(1)
    assert (255, 255, 255) == aci2rgb(7)


def test_0():
    with pytest.raises(IndexError):
        aci2rgb(0)


def test_256():
    with pytest.raises(IndexError):
        aci2rgb(256)


def test_luminance():
    black = luminance((0, 0, 0))
    blue = luminance((0, 0, 255))
    red = luminance((255, 0, 0))
    green = luminance((0, 255, 0))
    yellow = luminance((255, 255, 0))
    cyan = luminance((0, 255, 255))
    magenta = luminance((255, 0, 255))
    white = luminance((255, 255, 255))

    assert black == 0.0
    assert white == 1.0
    assert black < blue < red < magenta < green < cyan < yellow < white


class TestRGB:
    def test_from_hex_(self):
        assert RGB.from_hex("#aabbcc") == (0xAA, 0xBB, 0xCC)

    def test_from_hex_with_alpha(self):
        assert RGB.from_hex("#aabbccdd") == (0xAA, 0xBB, 0xCC)

    def test_from_floats(self):
        assert RGB.from_floats((1, 0.5, 0)) == (255, 128, 0)

    def test_to_floats(self):
        assert RGB(255, 128, 0).to_floats() == (1.0, 0.5019607843137255, 0.0)

    def test_to_hex(self):
        assert RGB.from_hex("#aabbcc").to_hex() == "#aabbcc"

    def test_luminance(self):
        rgb = RGB.from_hex("#aabbcc")
        assert rgb.luminance == luminance(rgb)


class TestRGBA:
    def test_from_hex(self):
        assert RGBA.from_hex("#aabbccdd") == (0xAA, 0xBB, 0xCC, 0xDD)

    def test_to_hex(self):
        assert RGBA.from_hex("#aabbccdd").to_hex() == "#aabbccdd"

    def test_from_floats(self):
        assert RGBA.from_floats((1, 0.5, 0, 0)) == (255, 128, 0, 0)
        assert RGBA.from_floats((1, 0.5, 0, 1)) == (255, 128, 0, 255)

    def test_to_floats(self):
        assert RGBA(255, 128, 0, 255).to_floats() == (1.0, 0.5019607843137255, 0.0, 1.0)

    def test_luminance(self):
        rgba = RGBA.from_hex("#aabbcc")
        assert rgba.luminance == luminance(rgba)

    def test_default_alpha_channel_is_opaque(self):
        assert RGBA(0xAA, 0xBB, 0xCC) == (0xAA, 0xBB, 0xCC, 0xFF)
        assert RGBA.from_hex("#aabbcc") == (0xAA, 0xBB, 0xCC, 0xFF)
        assert RGBA.from_floats((1, 0.5, 0)) == (255, 128, 0, 255)


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (C) 2019, Manfred Moitzi
# License: MIT License
from ezdxf.tools import set_flag_state


def test_set_flag_state():
    assert set_flag_state(0, 1, True) == 1
    assert set_flag_state(0b10, 1, True) == 0b11
    assert set_flag_state(0b111, 0b010, False) == 0b101
    assert set_flag_state(0b010, 0b111, True) == 0b111
    assert set_flag_state(0b1111, 0b1001, False) == 0b0110



# Copyright (c) 2019-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Any,
    Union,
    cast,
    Optional,
)
import logging

from ezdxf.math import Vec2, UVec
from ezdxf.entitydb import EntitySpace
from ezdxf.lldxf import const
from ezdxf.lldxf.validator import make_table_key
from .base import BaseLayout

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.layouts import BlockLayout
    from ezdxf.entities import GeoData, Viewport, DXFLayout, DXFGraphic, BlockRecord

logger = logging.getLogger("ezdxf")


def get_block_entity_space(doc: Drawing, block_record_handle: str) -> EntitySpace:
    block_record = doc.entitydb[block_record_handle]
    return block_record.entity_space  # type: ignore


class Layout(BaseLayout):
    """
    Layout representation - base class for :class:`Modelspace` and
    :class:`Paperspace`

    Every layout consist of a LAYOUT entity in the OBJECTS section, an
    associated BLOCK in the BLOCKS section and a BLOCK_RECORD_TABLE entry.

    layout_key: handle of the BLOCK_RECORD, every layout entity has this
        handle as owner attribute (entity.dxf.owner)

    There are 3 different layout types:

        1. :class:`Modelspace`
        2. active :class:`Paperspace` layout
        3. inactive :class:`Paperspace` layout

    Internal Structure

    For every layout exist a :class:`BlockLayout` object in
    :class:`BlocksSection` and a :class:`Layout` object
    (as :class:`Modelspace` or :class:`Paperspace`) in :class:`Layouts`.

    The entity space of the :class:`BlockLayout` object and the entity space
    of the :class:`Layout` object are the same object.

    """

    # plot_layout_flags of LAYOUT entity
    PLOT_VIEWPORT_BORDERS = 1
    SHOW_PLOT_STYLES = 2
    PLOT_CENTERED = 4
    PLOT_HIDDEN = 8
    USE_STANDARD_SCALE = 16
    PLOT_PLOTSTYLES = 32
    SCALE_LINEWEIGHTS = 64
    PRINT_LINEWEIGHTS = 128
    DRAW_VIEWPORTS_FIRST = 512
    MODEL_TYPE = 1024
    UPDATE_PAPER = 2048
    ZOOM_TO_PAPER_ON_UPDATE = 4096
    INITIALIZING = 8192
    PREV_PLOT_INIT = 16384

    def __init__(self, layout: DXFLayout, doc: Drawing):
        self.dxf_layout = layout
        handle = layout.dxf.get("block_record_handle", "0")
        try:
            block_record = doc.entitydb[handle]
        except KeyError:
            block_record = _find_layout_block_record(layout)  # type: ignore
            if block_record is None:
                raise const.DXFStructureError(
                    f"required BLOCK_RECORD #{handle} for layout '{layout.dxf.name}' "
                    f"does not exist"
                )
        if block_record.dxftype() != "BLOCK_RECORD":
            raise const.DXFStructureError(
                f"expected BLOCK_RECORD(#{handle}) for layout '{layout.dxf.name}' "
                f"has invalid entity type: {block_record.dxftype()}"
            )
        # link is maybe broken
        block_record.dxf.layout = layout.dxf.handle
        super().__init__(block_record)  # type: ignore

    @classmethod
    def new(cls, name: str, block_name: str, doc: Drawing, dxfattribs=None) -> Layout:
        """Returns the required structures for a new layout:

            - a :class:`BlockLayout` with BLOCK_RECORD, BLOCK and ENDBLK entities
            - LAYOUT entity in the objects section

        Args:
            name: layout name as shown in tabs of CAD applications e.g. 'Layout2'
            block_name: layout block name e.g. '*Paper_Space2'
            doc: drawing document
            dxfattribs: additional DXF attributes for LAYOUT entity

        (internal API)

        """
        block_layout: BlockLayout = doc.blocks.new(block_name)
        dxfattribs = dxfattribs or {}
        dxfattribs.update(
            {
                "name": name,
                "block_record_handle": block_layout.block_record_handle,
            }
        )
        dxf_layout = doc.objects.new_entity("LAYOUT", dxfattribs=dxfattribs)
        return cls(dxf_layout, doc)  # type: ignore

    @classmethod
    def load(cls, layout: DXFLayout, doc: Drawing):
        """Loading interface. (internal API)"""
        _layout = cls(layout, doc)
        _layout._repair_owner_tags()
        return _layout

    @property
    def name(self) -> str:
        """Layout name as shown in tabs of :term:`CAD` applications."""
        return self.dxf_layout.dxf.name

    @property  # dynamic DXF attribute dispatching, e.g. DXFLayout.dxf.layout_flags
    def dxf(self) -> Any:
        """Returns the DXF name space attribute of the associated
        :class:`~ezdxf.entities.DXFLayout` object.

        This enables direct access to the underlying LAYOUT entity,
        e.g. ``Layout.dxf.layout_flags``

        """
        return self.dxf_layout.dxf

    @property
    def block_record_name(self) -> str:
        """Returns the name of the associated BLOCK_RECORD as string."""
        return self.block_record.dxf.name

    def _repair_owner_tags(self) -> None:
        """Set `owner` and `paperspace` attributes of entities hosted by this
        layout to correct values.
        """
        layout_key = self.layout_key
        paperspace = 0 if self.is_modelspace else 1
        for entity in self:
            if entity.dxf.owner != layout_key:
                entity.dxf.owner = layout_key
            if entity.dxf.paperspace != paperspace:
                entity.dxf.paperspace = paperspace

    def __contains__(self, entity: Union[DXFGraphic, str]) -> bool:
        """Returns ``True`` if `entity` is stored in this layout.

        Args:
             entity: :class:`DXFGraphic` object or handle as hex string

        """
        if isinstance(entity, str):  # entity is a handle string
            entity = self.entitydb[entity]  # type: ignore
        return entity.dxf.owner == self.layout_key  # type: ignore

    def destroy(self) -> None:
        """Delete all entities and the layout itself from entity database and
        all linked structures.

        (internal API)
        """

        self.doc.objects.delete_entity(self.dxf_layout)
        # super() deletes block_record and associated entity space
        super().destroy()

    @property
    def plot_layout_flags(self) -> int:
        return self.dxf_layout.dxf.plot_layout_flags

    def reset_extents(
        self, extmin=(+1e20, +1e20, +1e20), extmax=(-1e20, -1e20, -1e20)
    ) -> None:
        """Reset `extents`_ to given values or the AutoCAD default values.

        "Drawing extents are the bounds of the area occupied by objects."
        (Quote Autodesk Knowledge Network)

        Args:
             extmin: minimum extents or (+1e20, +1e20, +1e20) as default value
             extmax: maximum extents or (-1e20, -1e20, -1e20) as default value

        """
        dxf = self.dxf_layout.dxf
        dxf.extmin = extmin
        dxf.extmax = extmax

    def reset_limits(self, limmin=None, limmax=None) -> None:
        """Reset `limits`_ to given values or the AutoCAD default values.

        "Sets an invisible rectangular boundary in the drawing area that can
        limit the grid display and limit clicking or entering point locations."
        (Quote Autodesk Knowledge Network)

        The :class:`Paperspace` class has an additional method
        :meth:`~Paperspace.reset_paper_limits` to deduce the default limits from
        the paper size settings.

        Args:
             limmin: minimum limits or (0, 0) as default
             limmax: maximum limits or (paper width, paper height) as default value

        """
        dxf = self.dxf_layout.dxf
        if limmin is None:
            limmin = (0, 0)
        if limmax is None:
            limmax = (dxf.paper_width, dxf.paper_height)

        dxf.limmin = limmin
        dxf.limmax = limmax

    def set_plot_type(self, value: int = 5) -> None:
        """
        === ============================================================
        0   last screen display
        1   drawing extents
        2   drawing limits
        3   view specific (defined by :attr:`Layout.dxf.plot_view_name`)
        4   window specific (defined by :meth:`Layout.set_plot_window_limits`)
        5   layout information (default)
        === ============================================================

        Args:
            value:  plot type

        Raises:
            DXFValueError: for `value` out of range

        """
        if 0 <= int(value) <= 5:
            self.dxf.plot_type = value
        else:
            raise const.DXFValueError("Plot type value out of range (0-5).")

    def set_plot_style(self, name: str = "ezdxf.ctb", show: bool = False) -> None:
        """Set plot style file of type `.ctb`.

        Args:
            name: plot style filename
            show: show plot style effect in preview? (AutoCAD specific attribute)

        """
        self.dxf_layout.dxf.current_style_sheet = name
        self.use_plot_styles(True)
        self.show_plot_styles(show)

    def get_plot_style_filename(self) -> str:
        return self.dxf_layout.dxf.current_style_sheet

    def set_plot_window(
        self,
        lower_left: tuple[float, float] = (0, 0),
        upper_right: tuple[float, float] = (0, 0),
    ) -> None:
        """Set plot window size in (scaled) paper space units.

        Args:
            lower_left: lower left corner as 2D point
            upper_right: upper right corner as 2D point

        """
        x1, y1 = lower_left
        x2, y2 = upper_right
        dxf = self.dxf_layout.dxf
        dxf.plot_window_x1 = x1
        dxf.plot_window_y1 = y1
        dxf.plot_window_x2 = x2
        dxf.plot_window_y2 = y2
        self.set_plot_type(4)

    def get_plot_unit_scale_factor(self) -> float:
        denom = self.dxf.scale_denominator
        numerator = self.dxf.scale_numerator
        scale = 1.0
        if denom:
            scale = numerator / denom
        if self.dxf.plot_paper_units == 1:
            return scale  # mm
        else:
            return scale * 25.4  # inch

    # plot layout flags setter
    def plot_viewport_borders(self, state: bool = True) -> None:
        self.set_plot_flags(self.PLOT_VIEWPORT_BORDERS, state)

    def show_plot_styles(self, state: bool = True) -> None:
        self.set_plot_flags(self.SHOW_PLOT_STYLES, state)

    def plot_centered(self, state: bool = True) -> None:
        self.set_plot_flags(self.PLOT_CENTERED, state)

    def plot_hidden(self, state: bool = True) -> None:
        self.set_plot_flags(self.PLOT_HIDDEN, state)

    def use_standard_scale(self, state: bool = True) -> None:
        self.set_plot_flags(self.USE_STANDARD_SCALE, state)

    def use_plot_styles(self, state: bool = True) -> None:
        self.set_plot_flags(self.PLOT_PLOTSTYLES, state)

    def scale_lineweights(self, state: bool = True) -> None:
        self.set_plot_flags(self.SCALE_LINEWEIGHTS, state)

    def print_lineweights(self, state: bool = True) -> None:
        self.set_plot_flags(self.PRINT_LINEWEIGHTS, state)

    def draw_viewports_first(self, state: bool = True) -> None:
        self.set_plot_flags(self.DRAW_VIEWPORTS_FIRST, state)

    def model_type(self, state: bool = True) -> None:
        self.set_plot_flags(self.MODEL_TYPE, state)

    def update_paper(self, state: bool = True) -> None:
        self.set_plot_flags(self.UPDATE_PAPER, state)

    def zoom_to_paper_on_update(self, state: bool = True) -> None:
        self.set_plot_flags(self.ZOOM_TO_PAPER_ON_UPDATE, state)

    def plot_flags_initializing(self, state: bool = True) -> None:
        self.set_plot_flags(self.INITIALIZING, state)

    def prev_plot_init(self, state: bool = True) -> None:
        self.set_plot_flags(self.PREV_PLOT_INIT, state)

    def set_plot_flags(self, flag, state: bool = True) -> None:
        self.dxf_layout.set_flag_state(flag, state=state, name="plot_layout_flags")


class Modelspace(Layout):
    """:class:`Modelspace` - not deletable, all entities of this layout are
    stored in the ENTITIES section of the DXF file, the associated
    "*Model_Space" block is empty, block name is fixed as "*Model_Space",
    the name is fixed as "Model".

    """

    @property
    def name(self) -> str:
        """Name of modelspace is fixed as "Model"."""
        return "Model"

    def new_geodata(self, dxfattribs=None) -> GeoData:
        """Creates a new :class:`GeoData` entity and replaces existing ones.
        The GEODATA entity resides in the OBJECTS section and not in the
        modelspace, it is linked to the modelspace by an
        :class:`~ezdxf.entities.ExtensionDict` located in BLOCK_RECORD of the
        modelspace.

        The GEODATA entity requires DXF R2010. The DXF reference does not
        document if other layouts than the modelspace supports geo referencing,
        so I assume getting/setting geo data may only make sense for the
        modelspace.

        Args:
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.GeoData` entity

        """
        if self.doc.dxfversion < const.DXF2010:
            raise const.DXFValueError("GEODATA entity requires DXF R2010 or later.")

        if dxfattribs is None:
            dxfattribs = {}
        xdict = self.get_extension_dict()
        geodata = self.doc.objects.add_geodata(
            owner=xdict.dictionary.dxf.handle,
            dxfattribs=dxfattribs,
        )
        xdict["ACAD_GEOGRAPHICDATA"] = geodata
        return geodata

    def get_geodata(self) -> Optional[GeoData]:
        """Returns the :class:`~ezdxf.entities.GeoData` entity associated to
        the modelspace or ``None``.
        """
        try:
            xdict = self.block_record.get_extension_dict()
        except AttributeError:
            return None
        try:
            return xdict["ACAD_GEOGRAPHICDATA"]
        except const.DXFKeyError:
            return None


class Paperspace(Layout):
    """There are two kind of paperspace layouts:

    1. Active Layout - all entities of this layout are stored in the ENTITIES
       section, the associated "*Paper_Space" block is empty, block name
       "*Paper_Space" is mandatory and also marks the active layout, the layout
       name can be an arbitrary string.

    2. Inactive Layout - all entities of this layouts are stored in the
       associated BLOCK called "*Paper_SpaceN", where "N" is an arbitrary
       number, I don't know if the block name schema "*Paper_SpaceN" is
       mandatory, the layout name can be an arbitrary string.

    There is no different handling for active layouts and inactive layouts in
    `ezdxf`, this differentiation is just for AutoCAD important and it is not
    documented in the DXF reference.

    """

    def rename(self, name: str) -> None:
        """Rename layout to `name`, changes the name displayed in tabs by
        CAD applications, not the internal BLOCK name. (internal API)

        Use method :meth:`~ezdxf.layouts.Layouts.rename` of the
        :meth:`~ezdxf.layouts.Layouts` class to rename paper space
        layouts.

        """
        self.dxf_layout.dxf.name = name

    def viewports(self) -> list[Viewport]:
        """Get all VIEWPORT entities defined in this paperspace layout."""
        return [e for e in self if e.is_alive and e.dxftype() == "VIEWPORT"]  # type: ignore

    def main_viewport(self) -> Optional[Viewport]:
        """Returns the main viewport of this paper space layout, or ``None``
        if no main viewport exist.

        """
        # Theory: the first VP found is the main VP of the layout, the attributes status
        # and id are ignored by BricsCAD and AutoCAD!?
        for viewport in self.viewports():
            dxf = viewport.dxf
            if dxf.hasattr("status") and dxf.status == 1:
                return viewport
            if dxf.id == 1:
                return viewport
        return None

    def add_viewport(
        self,
        center: UVec,
        size: tuple[float, float],
        view_center_point: UVec,
        view_height: float,
        status: int = 2,
        dxfattribs=None,
    ) -> Viewport:
        """Add a new :class:`~ezdxf.entities.Viewport` entity.

        Viewport :attr:`status`:

            - -1 is on, but is fully off-screen, or is one of the viewports that is not
              active because the $MAXACTVP count is currently being exceeded.
            - 0 is off
            - any value>0 is on and active. The value indicates the order of
              stacking for the viewports, where 1 is the "active viewport", 2 is the
              next, ...

        """
        dxfattribs = dxfattribs or {}
        width, height = size
        attribs = {
            "center": center,  # center in paperspace
            "width": width,  # width in paperspace
            "height": height,  # height in paperspace
            "status": status,
            "layer": "VIEWPORTS",
            # use separated layer to turn off for plotting
            "view_center_point": view_center_point,  # in modelspace
            "view_height": view_height,  # in modelspace
        }
        attribs.update(dxfattribs)
        viewport = cast("Viewport", self.new_entity("VIEWPORT", attribs))
        viewport.dxf.id = self.get_next_viewport_id()
        return viewport

    def get_next_viewport_id(self):
        viewports = self.viewports()
        if viewports:
            return max(vp.dxf.id for vp in viewports) + 1
        return 2

    def reset_viewports(self) -> None:
        """Delete all existing viewports, and create a new main viewport."""
        # remove existing viewports
        for viewport in self.viewports():
            self.delete_entity(viewport)
        self.add_new_main_viewport()

    def reset_main_viewport(self, center: UVec = None, size: UVec = None) -> Viewport:
        """Reset the main viewport of this paper space layout to the given
        values, or reset them to the default values, deduced from the paper
        settings. Creates a new main viewport if none exist.

        Ezdxf does not create a main viewport by default, because CAD
        applications don't require one.

        Args:
            center: center of the viewport in paper space units
            size: viewport size as (width, height) tuple in paper space units

        """
        viewport = self.main_viewport()
        if viewport is None:
            viewport = self.add_new_main_viewport()
        default_center, default_size = self.default_viewport_config()
        if center is None:
            center = default_center
        if size is None:
            size = default_size

        viewport.dxf.center = center
        width, height = size
        viewport.dxf.width = width
        viewport.dxf.height = height
        return viewport

    def default_viewport_config(
        self,
    ) -> tuple[tuple[float, float], tuple[float, float]]:
        dxf = self.dxf_layout.dxf
        if dxf.plot_paper_units == 0:  # inches
            unit_factor = 25.4
        else:  # mm
            unit_factor = 1.0

        # all paper parameters in mm!
        # all viewport parameters in paper space units inch/mm + scale factor!
        scale_factor = dxf.scale_denominator / dxf.scale_numerator

        def paper_units(value):
            return value / unit_factor * scale_factor

        paper_width = paper_units(dxf.paper_width)
        paper_height = paper_units(dxf.paper_height)
        # plot origin offset
        x_offset = paper_units(dxf.plot_origin_x_offset)
        y_offset = paper_units(dxf.plot_origin_y_offset)

        # printing area
        printable_width = (
            paper_width - paper_units(dxf.left_margin) - paper_units(dxf.right_margin)
        )
        printable_height = (
            paper_height - paper_units(dxf.bottom_margin) - paper_units(dxf.top_margin)
        )

        # AutoCAD viewport (window) size
        vp_width = paper_width * 1.1
        vp_height = paper_height * 1.1

        # center of printing area
        center = (
            printable_width / 2 - x_offset,
            printable_height / 2 - y_offset,
        )
        return center, (vp_width, vp_height)

    def add_new_main_viewport(self) -> Viewport:
        """Add a new main viewport."""
        center, size = self.default_viewport_config()
        vp_height = size[1]
        # create 'main' viewport
        main_viewport = self.add_viewport(
            center=center,  # no influence to 'main' viewport?
            size=size,  # I don't get it, just use paper size!
            view_center_point=center,  # same as center
            view_height=vp_height,  # view height in paper space units
            status=1,  # main viewport
        )
        if len(self.entity_space) > 1:
            # move main viewport to index 0 of entity space
            _vp = self.entity_space.pop()
            assert _vp is main_viewport
            self.entity_space.insert(0, main_viewport)

        main_viewport.dxf.id = 1  # set as main viewport
        main_viewport.dxf.flags = 557088  # AutoCAD default value
        self.set_current_viewport_handle(main_viewport.dxf.handle)
        return main_viewport

    def set_current_viewport_handle(self, handle: str) -> None:
        self.dxf_layout.dxf.viewport_handle = handle

    def page_setup(
        self,
        size: tuple[float, float] = (297, 210),
        margins: tuple[float, float, float, float] = (0, 0, 0, 0),
        units: str = "mm",
        offset: tuple[float, float] = (0, 0),
        rotation: int = 0,
        scale: Union[int, tuple[float, float]] = 16,
        name: str = "ezdxf",
        device: str = "DWG to PDF.pc3",
    ) -> None:
        """Setup plot settings and paper size and reset viewports.
        All parameters in given `units` (mm or inch).

        Reset paper limits, extents and viewports.

        Args:
            size: paper size as (width, height) tuple
            margins: (top, right, bottom, left) hint: clockwise
            units: "mm" or "inch"
            offset: plot origin offset is 2D point
            rotation: see table Rotation
            scale: integer in range [0, 32] defines a standard scale type or
                as tuple(numerator, denominator) e.g. (1, 50) for scale 1:50
            name: paper name prefix "{name}_({width}_x_{height}_{unit})"
            device: device .pc3 configuration file or system printer name

        === ============
        int Rotation
        === ============
        0   no rotation
        1   90 degrees counter-clockwise
        2   upside-down
        3   90 degrees clockwise
        === ============

        """
        if int(rotation) not in (0, 1, 2, 3):
            raise const.DXFValueError("valid rotation values: 0-3")

        if isinstance(scale, tuple):
            standard_scale = 16
            scale_num, scale_denom = scale
        elif isinstance(scale, int):
            standard_scale = scale
            scale_num, scale_denom = const.STD_SCALES.get(standard_scale, (1.0, 1.0))
        else:
            raise const.DXFTypeError(
                "Scale has to be an int or a tuple(numerator, denominator)"
            )
        if scale_num == 0:
            raise const.DXFValueError("Scale numerator can't be 0.")
        if scale_denom == 0:
            raise const.DXFValueError("Scale denominator can't be 0.")
        paper_width, paper_height = size
        margin_top, margin_right, margin_bottom, margin_left = margins
        units = units.lower()
        if units.startswith("inch"):
            units = "Inches"
            plot_paper_units = 0
            unit_factor = 25.4  # inch to mm
        elif units == "mm":
            units = "MM"
            plot_paper_units = 1
            unit_factor = 1.0
        else:
            raise const.DXFValueError('Supported units: "mm" and "inch"')

        # Setup PLOTSETTINGS
        # all paper sizes in mm
        dxf = self.dxf_layout.dxf
        if not dxf.hasattr("plot_layout_flags"):
            dxf.plot_layout_flags = dxf.get_default("plot_layout_flags")
        self.use_standard_scale(False)  # works best, don't know why
        dxf.page_setup_name = ""
        dxf.plot_configuration_file = device
        dxf.paper_size = f"{name}_({paper_width:.2f}_x_{paper_height:.2f}_{units})"
        dxf.left_margin = margin_left * unit_factor
        dxf.bottom_margin = margin_bottom * unit_factor
        dxf.right_margin = margin_right * unit_factor
        dxf.top_margin = margin_top * unit_factor
        dxf.paper_width = paper_width * unit_factor
        dxf.paper_height = paper_height * unit_factor
        dxf.scale_numerator = scale_num
        dxf.scale_denominator = scale_denom
        dxf.plot_paper_units = plot_paper_units
        dxf.plot_rotation = rotation

        x_offset, y_offset = offset
        dxf.plot_origin_x_offset = x_offset * unit_factor  # conversion to mm
        dxf.plot_origin_y_offset = y_offset * unit_factor  # conversion to mm
        dxf.standard_scale_type = standard_scale
        dxf.unit_factor = 1.0 / unit_factor  # 1/1 for mm; 1/25.4 ... for inch

        # Setup Layout
        self.reset_paper_limits()
        self.reset_extents()
        self.reset_viewports()

    def reset_paper_limits(self) -> None:
        """Set paper limits to default values, all values in paperspace units
        but without plot scale (?).

        """
        dxf = self.dxf_layout.dxf
        if dxf.plot_paper_units == 0:  # inch
            unit_factor = 25.4
        else:  # mm
            unit_factor = 1.0

        # all paper sizes are stored in mm
        paper_width = dxf.paper_width / unit_factor  # in plot paper units
        paper_height = dxf.paper_height / unit_factor  # in plot paper units
        left_margin = dxf.left_margin / unit_factor
        bottom_margin = dxf.bottom_margin / unit_factor
        x_offset = dxf.plot_origin_x_offset / unit_factor
        y_offset = dxf.plot_origin_y_offset / unit_factor
        # plot origin is the lower left corner of the printable paper area
        # limits are the paper borders relative to the plot origin
        shift_x = left_margin + x_offset
        shift_y = bottom_margin + y_offset
        dxf.limmin = (-shift_x, -shift_y)  # paper space units
        dxf.limmax = (paper_width - shift_x, paper_height - shift_y)

    def get_paper_limits(self) -> tuple[Vec2, Vec2]:
        """Returns paper limits in plot paper units, relative to the plot origin.

        plot origin = lower left corner of printable area + plot origin offset

        Returns:
            tuple (Vec2(x1, y1), Vec2(x2, y2)), lower left corner is (x1, y1),
            upper right corner is (x2, y2).

        """
        return Vec2(self.dxf.limmin), Vec2(self.dxf.limmax)

    def page_setup_r12(
        self,
        size: tuple[float, float] = (297, 210),
        margins: tuple[float, float, float, float] = (0, 0, 0, 0),
        units: str = "mm",
        offset: tuple[float, float] = (0, 0),
        rotation: float = 0,
        scale: Union[int, tuple[float, float]] = 16,
    ) -> None:
        # remove existing viewports
        for viewport in self.viewports():
            self.delete_entity(viewport)

        if int(rotation) not in (0, 1, 2, 3):
            raise const.DXFValueError("Valid rotation values: 0-3")

        if isinstance(scale, int):
            scale_num, scale_denom = const.STD_SCALES.get(scale, (1, 1))
        else:
            scale_num, scale_denom = scale

        if scale_num == 0:
            raise const.DXFValueError("Scale numerator can't be 0.")
        if scale_denom == 0:
            raise const.DXFValueError("Scale denominator can't be 0.")

        scale_factor = scale_denom / scale_num

        # TODO: don't know how to set inch or mm mode in R12
        units = units.lower()
        if units.startswith("inch"):
            units = "Inches"
            plot_paper_units = 0
            unit_factor = 25.4  # inch to mm
        elif units == "mm":
            units = "MM"
            plot_paper_units = 1
            unit_factor = 1.0
        else:
            raise const.DXFValueError('Supported units: "mm" and "inch"')

        # all viewport parameters are scaled paper space units
        def paper_units(value):
            return value * scale_factor

        margin_top = paper_units(margins[0])
        margin_right = paper_units(margins[1])
        margin_bottom = paper_units(margins[2])
        margin_left = paper_units(margins[3])
        paper_width = paper_units(size[0])
        paper_height = paper_units(size[1])
        self.doc.header["$PLIMMIN"] = (0, 0)
        self.doc.header["$PLIMMAX"] = (paper_width, paper_height)
        self.doc.header["$PEXTMIN"] = (0, 0, 0)
        self.doc.header["$PEXTMAX"] = (paper_width, paper_height, 0)

        # printing area
        printable_width = paper_width - margin_left - margin_right
        printable_height = paper_height - margin_bottom - margin_top

        # AutoCAD viewport (window) size
        vp_width = paper_width * 1.1
        vp_height = paper_height * 1.1

        # center of printing area
        center = (printable_width / 2, printable_height / 2)

        # create 'main' viewport
        main_viewport = self.add_viewport(
            center=center,  # no influence to 'main' viewport?
            size=(vp_width, vp_height),  # I don't get it, just use paper size!
            view_center_point=center,  # same as center
            view_height=vp_height,  # view height in paper space units
            status=1,  # main viewport
        )
        main_viewport.dxf.id = 1  # set as main viewport
        main_viewport.dxf.render_mode = 1000  # AutoDesk default (view mode?)

    def get_paper_limits_r12(self) -> tuple[Vec2, Vec2]:
        """Returns paper limits in plot paper units."""
        limmin = self.doc.header.get("$PLIMMIN", (0, 0))
        limmax = self.doc.header.get("$PLIMMAX", (0, 0))
        return Vec2(limmin), Vec2(limmax)


def _find_layout_block_record(layout: DXFLayout) -> BlockRecord | None:
    """Find and link the lost BLOCK_RECORD for the given LAYOUT entity."""

    def link_layout(block_record: BlockRecord) -> None:
        logger.info(
            f"fixing broken links for '{layout.dxf.name}' between {str(layout)} and {str(block_record)}"
        )
        layout.dxf.block_record_handle = block_record.dxf.handle
        block_record.dxf.layout = layout.dxf.handle

    doc = layout.doc
    assert doc is not None
    if layout.dxf.name == "Model":  # modelspace layout
        key = make_table_key("*Model_Space")
        for block_record in doc.tables.block_records:
            if make_table_key(block_record.dxf.name) == key:
                link_layout(block_record)
                return block_record
        return None

    # paperspace layout
    search_key = make_table_key("*Paper_Space")
    paper_space_records = [
        block_record
        for block_record in doc.tables.block_records
        if make_table_key(block_record.dxf.name).startswith(search_key)
    ]

    def search(key):
        for block_record in paper_space_records:
            layout_handle = block_record.dxf.get("layout", "0")
            if doc.entitydb.get(layout_handle) is key:
                link_layout(block_record)
                return block_record
        return None

    # first search all records for the lost record
    block_record = search(layout)
    if block_record is None:
        # link layout to the next orphaned record
        block_record = search(None)
    return block_record



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.addons.dxf2code import (
    entities_to_code,
    table_entries_to_code,
    block_to_code,
)
from ezdxf.addons.dxf2code import (
    _fmt_mapping,
    _fmt_list,
    _fmt_api_call,
    _fmt_dxf_tags,
)


import ezdxf.entities
from ezdxf.lldxf.types import dxftag
from ezdxf.lldxf.tags import Tags  # required by exec() or eval()
from ezdxf.entities.ltype import LinetypePattern  # required by exec() or eval()
from ezdxf.math import Vec3

doc = ezdxf.new("R2010")
msp = doc.modelspace()


def test_fmt_mapping():
    d = {"a": 1, "b": "str", "c": Vec3(), "d": "xxx \"yyy\" 'zzz'"}
    r = list(_fmt_mapping(d))
    assert r[0] == "'a': 1,"
    assert r[1] == "'b': \"str\","
    assert r[2] == "'c': (0.0, 0.0, 0.0),"
    assert r[3] == "'d': \"xxx \\\"yyy\\\" 'zzz'\","


def test_fmt_int_list():
    l = [1, 2, 3]
    r = list(_fmt_list(l))
    assert r[0] == "1,"
    assert r[1] == "2,"
    assert r[2] == "3,"


def test_fmt_float_list():
    l = [1.0, 2.0, 3.0]
    r = list(_fmt_list(l))
    assert r[0] == "1.0,"
    assert r[1] == "2.0,"
    assert r[2] == "3.0,"


def test_fmt_vector_list():
    from ezdxf.math import Vec3

    l = [Vec3(), (1.0, 2.0, 3.0)]
    r = list(_fmt_list(l))
    assert r[0] == "(0.0, 0.0, 0.0),"
    assert r[1] == "(1.0, 2.0, 3.0),"


def test_fmt_api_call():
    r = _fmt_api_call(
        "msp.add_line(",
        ["start", "end"],
        dxfattribs={"start": (0, 0), "end": (1, 0), "color": 7},
    )
    assert r[0] == "msp.add_line("
    assert r[1] == "    start=(0, 0),"
    assert r[2] == "    end=(1, 0),"
    assert r[3] == "    dxfattribs={"
    assert r[4] == "        'color': 7,"
    assert r[5] == "    },"
    assert r[6] == ")"


def test_fmt_dxf_tags():
    tags = [dxftag(1, "TEXT"), dxftag(10, (1, 2, 3))]
    code = "[{}]".format("".join(_fmt_dxf_tags(tags)))
    r = eval(code, globals())
    assert r == tags


def translate_to_code_and_execute(entity):
    code = str(entities_to_code([entity], layout="msp"))
    exec(code, globals())
    return msp[-1]


def test_line_to_code():
    from ezdxf.entities.line import Line

    entity = Line.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "start": (1, 2, 3),
            "end": (4, 5, 6),
        },
    )

    new_entity = translate_to_code_and_execute(entity)
    for name in ("color", "start", "end"):
        assert new_entity.get_dxf_attrib(name) == entity.get_dxf_attrib(name)


def test_point_to_code():
    from ezdxf.entities.point import Point

    entity = Point.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "location": (1, 2, 3),
        },
    )
    new_entity = translate_to_code_and_execute(entity)
    for name in ("color", "location"):
        assert new_entity.get_dxf_attrib(name) == entity.get_dxf_attrib(name)


def test_circle_to_code():
    from ezdxf.entities.circle import Circle

    entity = Circle.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "center": (1, 2, 3),
            "radius": 2,
        },
    )
    new_entity = translate_to_code_and_execute(entity)
    for name in ("color", "center", "radius"):
        assert new_entity.get_dxf_attrib(name) == entity.get_dxf_attrib(name)


def test_arc_to_code():
    from ezdxf.entities.arc import Arc

    entity = Arc.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "center": (1, 2, 3),
            "radius": 2,
            "start_angle": 30,
            "end_angle": 60,
        },
    )
    new_entity = translate_to_code_and_execute(entity)
    for name in ("color", "center", "radius", "start_angle", "end_angle"):
        assert new_entity.get_dxf_attrib(name) == entity.get_dxf_attrib(name)


def test_text_to_code():
    from ezdxf.entities.text import Text

    entity = Text.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "text": "xyz",
            "insert": (2, 3, 4),
        },
    )
    new_entity = translate_to_code_and_execute(entity)
    for name in ("color", "text", "insert"):
        assert new_entity.get_dxf_attrib(name) == entity.get_dxf_attrib(name)


def test_solid_to_code():
    from ezdxf.entities.solid import Solid

    entity = Solid.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "vtx0": (1, 2, 3),
            "vtx1": (4, 5, 6),
            "vtx2": (7, 8, 9),
            "vtx3": (3, 2, 1),
        },
    )
    new_entity = translate_to_code_and_execute(entity)
    for name in ("vtx0", "vtx1", "vtx2", "vtx3"):
        assert new_entity.get_dxf_attrib(name) == entity.get_dxf_attrib(name)


def test_shape_to_code():
    from ezdxf.entities.shape import Shape

    entity = Shape.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "name": "shape_name",
            "insert": (2, 3, 4),
        },
    )
    new_entity = translate_to_code_and_execute(entity)
    for name in ("color", "name", "insert"):
        assert new_entity.get_dxf_attrib(name) == entity.get_dxf_attrib(name)


def test_ellipse_to_code():
    from ezdxf.entities.ellipse import Ellipse

    entity = Ellipse.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "center": (1, 2, 3),
            "major_axis": (2, 0, 0),
            "ratio": 0.5,
            "start_param": 1,
            "end_param": 3,
        },
    )
    new_entity = translate_to_code_and_execute(entity)
    for name in (
        "color",
        "center",
        "major_axis",
        "ratio",
        "start_param",
        "end_param",
    ):
        assert new_entity.get_dxf_attrib(name) == entity.get_dxf_attrib(name)


def test_insert_to_code():
    from ezdxf.entities.insert import Insert

    entity = Insert.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "name": "block1",
            "insert": (2, 3, 4),
        },
    )
    new_entity = translate_to_code_and_execute(entity)
    for name in ("name", "insert"):
        assert new_entity.get_dxf_attrib(name) == entity.get_dxf_attrib(name)


def test_attdef_to_code():
    from ezdxf.entities.attrib import AttDef

    entity = AttDef.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "tag": "TAG1",
            "text": "Text1",
            "insert": (2, 3, 4),
        },
    )
    new_entity = translate_to_code_and_execute(entity)
    for name in ("tag", "text", "insert"):
        assert new_entity.get_dxf_attrib(name) == entity.get_dxf_attrib(name)


def test_mtext_to_code():
    from ezdxf.entities.mtext import MText

    entity = MText.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "insert": (2, 3, 4),
        },
    )
    text = "xxx \"yyy\" 'zzz'"
    entity.text = text
    new_entity = translate_to_code_and_execute(entity)
    for name in ("color", "insert"):
        assert new_entity.get_dxf_attrib(name) == entity.get_dxf_attrib(name)
    assert new_entity.text == "xxx \"yyy\" 'zzz'"


def test_lwpolyline_to_code():
    from ezdxf.entities.lwpolyline import LWPolyline

    entity = LWPolyline.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
        },
    )
    entity.set_points(
        [
            (1, 2, 0, 0, 0),
            (4, 3, 0, 0, 0),
            (7, 8, 0, 0, 0),
        ]
    )
    new_entity = translate_to_code_and_execute(entity)
    for name in ("color", "count"):
        assert new_entity.get_dxf_attrib(name) == entity.get_dxf_attrib(name)
    for np, ep in zip(new_entity.get_points(), entity.get_points()):
        assert np == ep


def test_polyline_to_code():
    # POLYLINE does not work without an entity space
    polyline = msp.add_polyline3d(
        [
            (1, 2, 3),
            (2, 3, 7),
            (9, 3, 1),
            (4, 4, 4),
            (0, 5, 8),
        ]
    )

    new_entity = translate_to_code_and_execute(polyline)
    # Are the last two entities POLYLINE entities?
    assert msp[-2].dxftype() == msp[-1].dxftype()
    assert len(new_entity) == len(polyline)
    assert new_entity.dxf.flags == polyline.dxf.flags
    for np, ep in zip(new_entity.points(), polyline.points()):
        assert np == ep


def cmp_vertices(a, b):
    return all(Vec3(v0).isclose(v1) for v0, v1 in zip(a, b))


def test_spline_to_code():
    from ezdxf.entities.spline import Spline

    entity = Spline.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "degree": 3,
        },
    )
    entity.fit_points = [(1, 2, 0), (4, 3, 0), (7, 8, 0)]
    entity.control_points = [(1, 2, 0), (4, 3, 0), (7, 8, 0)]
    entity.knots = [1, 2, 3, 4, 5, 6, 7]
    entity.weights = [1.0, 2.0, 3.0]
    new_entity = translate_to_code_and_execute(entity)
    for name in (
        "color",
        "n_knots",
        "n_control_points",
        "n_fit_points",
        "degree",
    ):
        assert new_entity.get_dxf_attrib(name) == entity.get_dxf_attrib(name)

    assert new_entity.knots == entity.knots
    assert cmp_vertices(new_entity.control_points, entity.control_points) is True
    assert cmp_vertices(new_entity.fit_points, entity.fit_points) is True
    assert new_entity.weights == entity.weights


def test_leader_to_code():
    from ezdxf.entities.leader import Leader

    entity = Leader.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
        },
    )
    entity.set_vertices(
        [
            (1, 2, 0),
            (4, 3, 0),
            (7, 8, 0),
        ]
    )
    new_entity = translate_to_code_and_execute(entity)
    assert new_entity.dxf.color == entity.dxf.color
    for np, ep in zip(new_entity.vertices, entity.vertices):
        assert np == ep


def test_mesh_to_code():
    from ezdxf.entities.mesh import Mesh
    from ezdxf.render.forms import cube

    entity = Mesh.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
        },
    )
    c = cube()
    entity.vertices = c.vertices
    entity.faces = c.faces

    assert len(entity.vertices) == 8
    new_entity = translate_to_code_and_execute(entity)
    assert cmp_vertices(entity.vertices, new_entity.vertices) is True
    assert list(entity.faces) == list(new_entity.faces)


def test_layer_entry():
    from ezdxf.entities.layer import Layer

    layer = Layer.new("LAYER", dxfattribs={"name": "TestTest", "color": 3})
    code = table_entries_to_code([layer], drawing="doc")
    exec(str(code), globals())
    layer = doc.layers.get("TestTest")
    assert layer.dxf.color == 3


def test_ltype_entry():
    from ezdxf.entities.ltype import Linetype

    ltype = Linetype.new(
        "FFFF",
        dxfattribs={
            "name": "TEST",
            "description": "TESTDESC",
        },
    )
    ltype.setup_pattern([0.2, 0.1, -0.1])
    code = table_entries_to_code([ltype], drawing="doc")
    exec(str(code), globals())
    new_ltype = doc.linetypes.get("TEST")
    assert new_ltype.dxf.description == ltype.dxf.description
    assert new_ltype.pattern_tags.tags == ltype.pattern_tags.tags
    # all imports added
    assert any(line.endswith("Tags") for line in code.imports)
    assert any(line.endswith("dxftag") for line in code.imports)
    assert any(line.endswith("LinetypePattern") for line in code.imports)


def test_block_to_code():
    testdoc = ezdxf.new()
    block = testdoc.blocks.new("TestBlock", dxfattribs={"description": "test"})
    block.add_line((1, 1), (2, 2))
    code = block_to_code(block, drawing="doc")
    exec(str(code), globals())
    new_block = doc.blocks.get("TestBlock")
    assert new_block.block.dxf.description == block.block.dxf.description
    assert new_block[0].dxftype() == block[0].dxftype()


def test_hatch_to_code():
    from ezdxf.entities import Hatch

    hatch = Hatch()
    hatch.set_pattern_fill(name="ANGLE")
    hatch.paths.add_polyline_path(
        [(0, 0), (100, 0), (100, 100), (0, 100)], is_closed=True
    )

    new_hatch = translate_to_code_and_execute(hatch)
    assert isinstance(new_hatch, Hatch)
    assert new_hatch.has_pattern_fill
    assert len(new_hatch.pattern.lines) == len(hatch.pattern.lines)


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest
import math
from ezdxf.math import (
    is_planar_face,
    Vec3,
    Vec2,
    subdivide_face,
    intersection_ray_ray_3d,
    intersection_line_line_3d,
    normal_vector_3p,
    safe_normal_vector,
    NULLVEC,
    X_AXIS,
    Y_AXIS,
    Z_AXIS,
    subdivide_ngons,
    distance_point_line_3d,
    best_fit_normal,
    Matrix44,
    BarycentricCoordinates,
    linear_vertex_spacing,
    is_vertex_order_ccw_3d,
)

from ezdxf.render.forms import square, circle

REGULAR_FACE = Vec3.list([(0, 0, 0), (1, 0, 1), (1, 1, 1), (0, 1, 0)])
IRREGULAR_FACE = Vec3.list([(0, 0, 0), (1, 0, 1), (1, 1, 0), (0, 1, 0)])
REGULAR_FACE_WRONG_ORDER = Vec3.list([(0, 0, 0), (1, 1, 1), (1, 0, 1), (0, 1, 0)])
ONLY_COLINEAR_EDGES = Vec3.list([(0, 0, 0), (1, 0, 0), (2, 0, 0), (3, 0, 0)])
REGULAR_FACE_WITH_COLINEAR_EDGE = Vec3.list(
    [(0, 0, 0), (1, 0, 0), (2, 0, 0), (3, 0, 0), (1.5, 2.0, 0)]
)


def test_face_count():
    assert is_planar_face(REGULAR_FACE[:3]) is True
    assert is_planar_face(REGULAR_FACE[:2]) is False


def test_regular_face():
    assert is_planar_face(REGULAR_FACE) is True


def test_irregular_face():
    assert is_planar_face(IRREGULAR_FACE) is False


def test_only_colinear_edges():
    assert is_planar_face(ONLY_COLINEAR_EDGES) is False


def test_regular_face_with_colinear_edge():
    assert is_planar_face(REGULAR_FACE) is True


def test_does_not_detect_wrong_order():
    assert is_planar_face(REGULAR_FACE_WRONG_ORDER) is True


def test_subdivide_square_in_quads():
    b = square(2)
    result = list(subdivide_face(b, quads=True))
    assert len(result) == 4
    assert result[0] == ((0, 0), (1, 0), (1, 1), (0, 1))


def test_subdivide_square_in_triangles():
    b = square(2)
    result = list(subdivide_face(b, quads=False))
    assert len(result) == 8
    assert result[0] == ((0, 1), (0, 0), (1, 1))
    assert result[1] == ((0, 0), (1, 0), (1, 1))


def test_subdivide_triangle():
    t = Vec3.list([(0, 0), (2, 0), (1, 2)])
    assert len(list(subdivide_face(t, quads=True))) == 3
    assert len(list(subdivide_face(t, quads=False))) == 6


def test_subdivide_ngons():
    hexagon = list(circle(6))
    result = list(subdivide_ngons([hexagon]))
    assert len(result) == 6


def test_subdivide_vec2_square_in_quads():
    b = Vec2.list(square(2))
    result = list(subdivide_face(b, quads=True))
    assert len(result) == 4
    assert result[0] == ((0, 0), (1, 0), (1, 1), (0, 1))


class TestIntersectionRayRay3d:
    @pytest.fixture
    def ray1(self):
        return Vec3(0, 0, 0), Vec3(1, 0, 0)

    @pytest.fixture
    def ray2(self):
        return Vec3(0, 0, 0), Vec3(0, 0, 1)

    def test_parallel_rays_return_empty_tuple(self, ray1, ray2):
        result = intersection_ray_ray_3d(ray1, ray1)
        assert len(result) == 0
        assert bool(result) is False

    def test_intersecting_rays_return_one_tuple(self, ray1, ray2):
        result = intersection_ray_ray_3d(ray1, ray2)
        assert len(result) == 1
        assert bool(result) is True
        assert result == (Vec3(0, 0, 0),)

    def test_not_intersecting_and_not_parallel_rays_return_two_tuple(self, ray1, ray2):
        line3 = (Vec3(0, 0, 1), Vec3(0, 1, 1))
        result = intersection_ray_ray_3d(ray1, line3)
        assert len(result) == 2
        assert bool(result) is True
        # returns points of closest approach on each ray
        assert Vec3(0, 0, 1) in result
        assert Vec3(0, 0, 0) in result

    def test_intersecting_rays(self):
        ray1 = (Vec3(1, 0, 0), Vec3(1, 1, 0))
        ray2 = (Vec3(0, 0.5, 0), Vec3(1, 0.5, 0))
        result = intersection_ray_ray_3d(ray1, ray2)
        assert len(result) == 1

    def test_random_intersecting_rays(self):
        for _ in range(5):
            intersection_point = Vec3.random(5)
            ray1 = (intersection_point, intersection_point + Vec3.random())
            ray2 = (intersection_point, intersection_point - Vec3.random())

            result = intersection_ray_ray_3d(ray1, ray2)
            assert len(result) == 1
            assert result[0].isclose(intersection_point)


class TestIntersectingLines3d:
    @pytest.fixture
    def line1(self):
        return Vec3(0, 0, 0), Vec3(2, 0, 0)

    @pytest.fixture
    def line2(self):
        return Vec3(1, -1, 0), Vec3(1, 1, 0)

    @pytest.fixture
    def line3(self):
        return Vec3(3, -1, 0), Vec3(3, 1, 0)

    @pytest.fixture
    def line4(self):
        return Vec3(2, -1, 0), Vec3(2, 1, 0)

    def test_real_intersecting_lines(self, line1, line2):
        assert intersection_line_line_3d(line1, line2, virtual=False).isclose((1, 0, 0))

    def test_virtual_intersecting_lines(self, line1, line3):
        assert intersection_line_line_3d(line1, line3, virtual=True).isclose((3, 0, 0))

    def test_not_intersecting_lines(self, line1, line3):
        assert intersection_line_line_3d(line1, line3, virtual=False) is None

    def test_touching_lines_do_intersect(self, line1, line4):
        assert intersection_line_line_3d(line1, line4, virtual=False).isclose((2, 0, 0))

    @pytest.mark.parametrize(
        "p2", [(4, 0), (0, 4), (4, 4)], ids=["horiz", "vert", "diag"]
    )
    def test_coincident_lines_do_not_intersect(self, p2):
        line = (Vec3(), Vec3(p2))
        assert intersection_line_line_3d(line, line, virtual=False) is None


RH_ORTHO = [
    (NULLVEC, X_AXIS, Y_AXIS, Z_AXIS),
    (NULLVEC, Y_AXIS, X_AXIS, -Z_AXIS),
    (NULLVEC, Z_AXIS, X_AXIS, Y_AXIS),
    (NULLVEC, X_AXIS, Z_AXIS, -Y_AXIS),
    (NULLVEC, Y_AXIS, Z_AXIS, X_AXIS),
    (NULLVEC, Z_AXIS, Y_AXIS, -X_AXIS),
]


@pytest.mark.parametrize("a,b,c,r", RH_ORTHO)
def test_normal_vector_for_3_points(a, b, c, r):
    assert normal_vector_3p(a, b, c) == r


def test_safe_normal_vector_regular():
    vertices = Vec3.list([(0, 0, 0), (1, 0, 0), (1, 1, 0)])
    assert safe_normal_vector(vertices).isclose((0, 0, 1))


def test_safe_normal_vector_for_coincident_vertices():
    vertices = Vec3.list([(0, 0, 0), (0, 0, 0), (1, 0, 0), (1, 1, 0)])
    assert safe_normal_vector(vertices).isclose((0, 0, 1))


def test_safe_normal_vector_for_colinear_vertices():
    vertices = Vec3.list([(0, 0, 0), (0.5, 0, 0), (1, 0, 0), (1, 1, 0)])
    assert safe_normal_vector(vertices).isclose((0, 0, 1))


def test_safe_normal_vector_raises_exception_for_undefined_normal_vector():
    vertices = Vec3.list([(0, 0, 0), (1, 0, 0), (2, 0, 0)])
    with pytest.raises(ZeroDivisionError):
        safe_normal_vector(vertices)


@pytest.mark.parametrize(
    "points, expected",
    [
        ([(10, 3), (0, 0), (1, 0)], 3),  # left of line
        ([(-10, 0), (0, 0), (1, 0)], 0),  # on line
        ([(2, -4), (0, 0), (1, 0)], 4),  # right of line
        ([(5, 0), (0, 5), (0, 2)], 5),
        ([(1, 0, 1), (1, 1, 1), (0, 0, 0)], 0.8164965809277259),
    ],
)
def test_distance_point_line_3d(points, expected):
    p, a, b = Vec3.generate(points)
    assert distance_point_line_3d(p, a, b) == pytest.approx(expected)


@pytest.mark.parametrize("x", [1e-99, 1e-9, 0, 1e9, 1e99])
def test_distance_point_line_3d_no_line(x):
    """Start point is equal or close to end point."""
    s = Vec3(1, 0, x)
    e = Vec3(1, 0, x)
    with pytest.raises(ZeroDivisionError):
        distance_point_line_3d(Vec3(1, 0, 0), s, e)


class TestBestFitNormal:
    @pytest.mark.parametrize("a,b,c,r", RH_ORTHO)
    def test_if_returns_right_handed_normals(self, a, b, c, r):
        assert best_fit_normal((a, b, c)) == r

    @pytest.fixture(scope="class")
    def vertices(self):
        return Vec3.list([(0, 0), (3, 0), (3, 4), (4, 8), (1, 5), (0, 2)])

    @pytest.fixture(scope="class")
    def matrix(self):
        return Matrix44.chain(
            Matrix44.x_rotate(0.75),
            Matrix44.translate(2, 3, 4),
        )

    def test_transformed_counter_clockwise_vertices_ccw(self, vertices, matrix):
        v = matrix.transform_vertices(vertices)
        normal = matrix.transform_direction(Z_AXIS)
        assert best_fit_normal(v).isclose(normal)

    def test_transformed_clockwise_vertices(self, vertices, matrix):
        v = matrix.transform_vertices(reversed(vertices))
        normal = matrix.transform_direction(-Z_AXIS)
        assert best_fit_normal(v).isclose(normal)


class TestBarycentricCoords:
    @pytest.fixture
    def bc(self):
        return BarycentricCoordinates((0, 0, 0), (5, 0, 0), (5, 4, 0))

    def test_basic_coords(self, bc):
        assert bc.from_cartesian(bc.a) == (1, 0, 0)
        assert bc.from_cartesian(bc.b) == (0, 1, 0)
        assert bc.from_cartesian(bc.c) == (0, 0, 1)

    def test_center_of_mass_property(self, bc):
        p = (bc.a + bc.b + bc.c) / 3
        b = bc.from_cartesian(p)
        assert b.isclose((1 / 3.0, 1 / 3.0, 1 / 3.0))

    @pytest.mark.parametrize("p", [(0, 4, 0), (0, -1, 0), (7, 0, 0)])
    def test_point_outside_triangle(self, bc, p):
        p = Vec3(p)
        b = bc.from_cartesian(p)
        assert any(b0 < 0 for b0 in b) is True
        assert sum(b) == pytest.approx(1.0)
        assert p.isclose(bc.to_cartesian(b))

    @pytest.mark.parametrize(
        "p",
        [
            # tests the normal projection of p onto (a, b, c)
            (4, 1, 0),
            (4, 1, 1),
            (4, 1, -1),
        ],
    )
    def test_point_inside_triangle(self, bc, p):
        b = bc.from_cartesian(p)
        assert all(0 <= b0 <= 1 for b0 in b) is True
        assert sum(b) == pytest.approx(1.0)


class TestLinearVertexSpacing:
    @pytest.mark.parametrize("count", [-1, 0, 1, 2, 3])
    def test_returns_always_two_or_more_vertices(self, count):
        assert len(linear_vertex_spacing(Vec3(), Vec3(1, 0), count)) >= 2

    def test_works_if_start_is_equal_to_end(self):
        assert len(linear_vertex_spacing(Vec3(), Vec3(), 5)) == 5

    @pytest.mark.parametrize("count", [2, 3, 4, 5])
    def test_correct_spacing_in_Q1(self, count):
        x = count - 1
        vertices = linear_vertex_spacing(Vec3(), Vec3(x, x, x), count)
        assert len(vertices) == count
        for x in range(count):
            assert vertices[x].isclose((x, x, x))

    @pytest.mark.parametrize("count", [2, 3, 4, 5])
    def test_correct_spacing_in_Q3(self, count):
        x = count - 1
        vertices = linear_vertex_spacing(Vec3(), Vec3(-x, -x, -x), count)
        assert len(vertices) == count
        for x in range(count):
            assert vertices[x].isclose((-x, -x, -x))


I_BEAM = Vec3.list(
    [
        (0, 0),
        (3, 0),
        (3, 1),
        (2, 1),
        (2, 2),
        (3, 2),
        (3, 3),
        (0, 3),
        (0, 2),
        (1, 2),
        (1, 1),
        (0, 1),
    ]
)


class TestIsVertexOrderCCW:
    def test_xy_plane(self):
        assert is_vertex_order_ccw_3d(I_BEAM, Vec3(0, 0, 1)) is True

    def test_xy_plane_inv(self):
        assert is_vertex_order_ccw_3d(I_BEAM, Vec3(0, 0, -1)) is False

    def test_yz_plane_up(self):
        m = Matrix44.x_rotate(math.pi / 2)
        vertices = list(m.transform_vertices(I_BEAM))
        assert is_vertex_order_ccw_3d(vertices, Vec3(0, 1, 0)) is True

    def test_yz_plane_inv(self):
        m = Matrix44.x_rotate(math.pi / 2)
        vertices = list(m.transform_vertices(I_BEAM))
        assert is_vertex_order_ccw_3d(vertices, Vec3(0, -1, 0)) is False

    def test_xz_plane_up(self):
        m = Matrix44.y_rotate(math.pi / 2)
        vertices = list(m.transform_vertices(I_BEAM))
        assert is_vertex_order_ccw_3d(vertices, Vec3(1, 0, 0)) is True

    def test_xz_plane_inv(self):
        m = Matrix44.y_rotate(math.pi / 2)
        vertices = list(m.transform_vertices(I_BEAM))
        assert is_vertex_order_ccw_3d(vertices, Vec3(-1, 0, 0)) is False

    def test_square_xy_plane(self):
        square = Vec3.list([(0, 0), (1, 0), (1, 1), (0, 1)])
        assert is_vertex_order_ccw_3d(square, Vec3(0, 0, 1)) is True
        assert is_vertex_order_ccw_3d(square, Vec3(0, 0, -1)) is False


if __name__ == "__main__":
    pytest.main([__file__])
